import numpy as _np


class TrinomialSpreadOption:
    __name__ = "TrinomialSpreadOption"
    __title__ = "Trinomial Tree Spread Option Model"

    def __init__(
        self,
        S1: float,
        S2: float,
        K: float,
        t: float,
        r: float,
        b: float,
        sigma1: float,
        sigma2: float,
        rho: float,
        type: str = "european",
        n: int = 5,
    ):
        self._S1 = S1
        self._S2 = S2
        self._K = K
        self._t = t
        self._r = r
        self._b = b
        self._sigma1 = sigma1
        self._sigma2 = sigma2
        self._rho = rho
        self._n = n
        self._type = type

    def call(self):
        return self._calc_price(z=1, type=self._type, tree=False)

    def _calc_price(self, z, type, tree):
        n = self._n
        S1 = self._S1
        sd1 = self._sigma1
        S2 = self._S2
        sd2 = self._sigma2
        rho = self._rho
        dt = self._t / n
        r = self._r
        b = self._b

        pm = _np.array([
            [1,4,1],
            [4,16,4],
            [1,4,1]
        ])/36

        sdp = sd1 * sd1 + sd2 * sd2
        # fmt: off
        lam1 = 0.5 * (sdp + _np.sqrt(sdp * sdp - 4 * (1 - rho * rho) * sd1 * sd1 * sd2 * sd2))
        lam2 = 0.5 * (sdp - _np.sqrt(sdp * sdp - 4 * (1 - rho * rho) * sd1 * sd1 * sd2 * sd2))
        the = _np.arctan((lam1-sd1*sd1)/(rho*sd1*sd2))
        Df = _np.exp(b-r*dt)
        h = _np.exp(-1*_np.sin(the)*_np.sqrt(lam2)* _np.sqrt(3 * dt) - 0.5*sd1*sd1*dt)
        v = _np.exp(_np.cos(the)*_np.sqrt(lam1)* _np.sqrt(3 * dt) - 0.5*sd1*sd1*dt)

        S1_T = S1*self._get_gauss(h,v,n)

        h = _np.exp(_np.cos(the)*_np.sqrt(lam2)* _np.sqrt(3 * dt) - 0.5*sd2*sd2*dt)
        v = _np.exp(_np.sin(the)*_np.sqrt(lam1)* _np.sqrt(3 * dt) - 0.5*sd2*sd2*dt)

        S2_T = S2*self._get_gauss(h,v,n)

        OptionValue = _np.maximum(0,((S1_T-S2_T)-self._K)*z)

        OptionValue = self._euro(OptionValue, n, Df, pm)

        OptionValue = _np.sum(_np.concatenate(_np.multiply(OptionValue[-1][n-1:n+2, n-1:n+2],pm)))*Df

        return OptionValue
        
    def _euro(self, OptionValue, n, Df, pm):
        tr = OptionValue.copy()
        old = tr.copy()

        tr = list()

        # step back in time from t=T to t=0 
        for j in _np.arange(1,n):

            # create new smaller matrix for expected value cale from passed, larger matrix
            latest = _np.zeros((2*(n-j)+1,2*(n-j)+1))
            latest = self._calc_nodes(old, latest, Df, pm)

            old = latest.copy()
            # create new matrix of same size as original matrix with new matrix
            # centered inside and stack
            latest = _np.pad(latest, (j, j))

            tr.append(latest)

        return tr

    def _calc_nodes(self, tr, ntr, Df, pm):
        # loop through elements 1 element away from the edge and calc
        # weighted sum
        for i in range(1,tr.shape[0]-1):
            for j in range(1,tr.shape[0]-1):
                ntr[i-1,j-1] = _np.sum(_np.concatenate(_np.multiply(tr[i-1:i+2, j-1:j+2],pm)))*Df
        return ntr





    def _get_gauss(self, h, v, n):
        """
        Calculate the joint innovation for one of the assets
        """
        # fmt: off
        OptionValue1 = _np.power(h, _np.arange(0, 2 * n + 1)) \
                            * (1 / h) ** _np.arange(2 * n, -1, -1)
        OptionValue1 = _np.broadcast_to(OptionValue1, (2 * n + 1, 2 * n + 1))

        OptionValue2 = _np.power(v, _np.arange(0, 2 * n + 1)) \
                            * (1 / v) ** _np.arange(2 * n, -1, -1)
        OptionValue2 = _np.broadcast_to(_np.flip(OptionValue2), (2 * n + 1, 2 * n + 1)).T
        # fmt: on

        OptionValue = OptionValue2 * OptionValue1

        return OptionValue


if __name__ == "__main__":

    opt = TrinomialSpreadOption(70, 60, 0, 1 / 12, 0.03, 0.03, 0.2, 0.1, 0.5, n=100)
    ret = opt._calc_price(1, type="euro", tree=False)

    print(ret)



from ..base import GreeksFDM, Option as _Option
from ..utils import docstring_from
from ..vanillaoptions import GBSOption as _GBSOption

from .mc_innovations import *
from .mc_paths import *
from .mc_payoffs import *

import numpy as _np
import warnings as _warnings


class MonteCarloOption:  # _Option
    """
    Class for the valuation of options using Monte Carlo methods using passed innovation, path and
    payoff calculation classes that inherit from the classes Innovations, Path and Payoff respectively.
    Parent classes are defined within this module. See example below.

    Parameters
    ----------
    mc_paths : int
        Number of payoff paths to generate per mc_loop. Total number of monte carlo sample paths
        is mc_loops * mc_paths.
    mc_loops : int
        Number of times to generate mc_path number of monte carlo paths. Any positive number.
        Total number of monte carlo sample paths is mc_loops * mc_paths.
    path_length : int
        Number of time steps to maturity t. dt = t/path_length.
    S : float
        Level or index price.
    K : float
        Strike price.
    t : float
        Time-to-maturity in fractional years. i.e. 1/12 for 1 month, 1/252 for 1 business day, 1.0 for 1 year.
    r : float
        Risk-free-rate in decimal format (i.e. 0.01 for 1%).
    b : float
        Annualized cost-of-carry rate, e.g. 0.1 means 10%.
    sigma : float
        Annualized volatility of the underlying asset. Optional if calculating implied volatility.
        Required otherwise. By default None.
    Innovation : Innovations class
        An Innovations subclass that has a sample_innovation method for generating innovations for
        the path generation step. sample_innovation method must return an numpy array of shape
        (mc_paths, path_length). See parent definition below.
    Path : Path subclass
        A Path subclass that has a generate_path method for generating monte carlo paths using the
        innovations of the Innovation class. generate_path method must return a numpy array of
        shape (mc_paths, path_length). See parent definition below.
    Payoff : Payoff sublass
        A Payoff subclass that has both the call and put methods for calculation strike payoffs
        at option maturity. Class must be able to calculate the payoffs based on a numpy array
        of monte carlo paths and call and put methods must return numpy arrays of shape
        (mc_paths,). See parent definition below.
    trace : bool
        If True, it will return the average option value per mc_loop as well as the cumulative
        average of these averages. Only useful with mc_loops > 1. By default False.
    antithetic : bool
        If True, innovations generated from the Innovations class is combined with it's negative
        version (i.e. -1 * Innovations) to center the innovations around zero. Note if set to
        True, this will double the number of mc_paths. By default False.
    standardization : bool
        If True, the first (mean) and second (standard deviation) moments of the generated
        innovations are made to match 0 and 1 respectively to account for errors in a pseudo
        random number generator. By default False.
    eps : numpy array
        pregenerated innovations to be used instead of one generated by an innovation class. Note
        that the innovation class must still be passed. By default None. For testing purposes only.

    Notes
    -----
    Including a cost of carry term b, the model can
    used to price European Options on:

    b = r       stocks (Black and Scholes’ stock option model)
    b = r - q   stocks and stock indexes paying a continuous dividend yield q (Merton’s stock option model)
    b = 0       futures (Black’s futures option model)
    b = r - rf  currency options with foreign interst rate rf (Garman and Kohlhagen’s currency option model)

    Parent Classes
    --------------

    class Innovations(ABC):
        def __init__(self, mc_paths, path_length, eps=None):
            self.mc_paths = mc_paths
            self.path_length = path_length
            self._eps = eps  # for testing only

        @abstractmethod
        def sample_innovation(self, init=False):
            pass # must return a numpy array of shape (mc_paths, path_length)

    class Path(ABC):
        def __init__(self, epsilon, sigma, dt, b):
            self.epsilon = epsilon # array of innovations
            self.sigma = sigma
            self.dt = dt # calculated dt from t/path_length
            self.b = b

        @abstractmethod
        def generate_path(self):
            pass # must return a numpy array of shape (mc_paths, path_length)

    class Payoff(ABC):
        def __init__(self, path, S: float, K: float, t: float, r: float, b: float, sigma: float):
            self.path = path
            self.S = S
            self.K = K
            self.t = t
            self.r = r
            self.b = b
            self.sigma = sigma

        @abstractmethod
        def call(self):
            pass # must return an average payoff for each mc_loop

        @abstractmethod
        def put(self):
            pass # must return an average payoff for each mc_loop

    Returns
    -------
    MonteCarloOption object.

    Example
    -------
    >>> import finoptions as fo
    >>> from scipy.stats import qmc, norm
    >>> import numpy as np
    >>> from finoptions.monte_carlo_options import Innovations, Path, Payoff
    >>> S = 100
    >>> K = 100
    >>> t = 1 / 12
    >>> sigma = 0.4
    >>> r = 0.10
    >>> b = 0.1
    >>> path_length = 30
    >>> mc_paths = 5000
    >>> mc_loops = 50

    >>> class NormalSobolInnovations(Innovations):
            def sample_innovation(self, scramble=True):
                sobol = qmc.Sobol(self.path_length, scramble=scramble).random(self.mc_paths)
                if scramble == False:
                    # add new sample since if not scrambled first row is zero which leads to -inf when normalized
                    sobol = sobol[1:]
                    sobol = np.append(
                        sobol,
                        qmc.Sobol(self.path_length, scramble=scramble).fast_forward(self.mc_paths).random(1),
                        axis=0,
                    )
                sobol = norm.ppf(sobol)
                return sobol

    >>> class WienerPath(Path):
            def generate_path(self, **kwargs):
                return (self.b - (self.sigma ** 2) / 2) * self.dt + self.sigma * np.sqrt(self.dt)  * self.epsilon

    >>> class PlainVanillaPayoff(Payoff):
            def call(self):
                St = self.S * np.exp(np.sum(self.path.generate_path(), axis=1))
                return np.exp(-self.r * self.t) * np.maximum(St - self.K, 0)

            def put(self):
                St = self.S * np.exp(np.sum(self.path.generate_path(), axis=1))
                return np.exp(-self.r * self.t) * np.maximum(self.K - St, 0)

    >>> opt = fo.monte_carlo_options.MonteCarloOption(
            mc_loops, path_length, mc_paths,
            S, K, t, r, b, sigma,
            NormalSobolInnovations, WienerPath, PlainVanillaPayoff,
            trace=False, antithetic=True, standardization=False
        )
    >>> print(opt.call().mean())
    >>> print(opt.put().mean())
    >>> opt.greeks(call=True) # not working yet

    References
    ----------
    [1] Haug E.G., The Complete Guide to Option Pricing Formulas

    """

    __name__ = "MCOption"
    __title__ = "Monte Carlo Simulation Option"

    def __init__(
        self,
        mc_loops: int,
        path_length: int,
        mc_paths: int,
        S: float,
        K: float,
        t: float,
        r: float,
        b: float,
        sigma: float = None,
        Innovation: Innovations = None,
        Path=None,
        Payoff=None,
        trace=False,
        antithetic=False,
        standardization=False,
        eps=None,
        **kwargs,
    ):

        self._mc_loops = mc_loops
        self._path_length = path_length
        self._mc_paths = mc_paths
        self._dt = t / path_length
        self._S = S
        self._K = K
        self._t = t
        self._r = r
        self._b = b
        self._sigma = sigma
        self._Innovation = Innovation(mc_paths, path_length, eps)
        self._Path = Path
        self._Payoff = Payoff
        self._trace = trace
        self._antithetic = antithetic
        self._standardization = standardization
        self._eps = eps
        self._kwargs = kwargs

    def call(self):
        """
        Returns an array of the average option call price per Monte Carlo Loop (mc_loop). Final option value
        is the average of the returned array.

        Returns
        -------
        numpy array

        Example
        -------
        >>> import finoptions as fo
        >>> inno = fo.monte_carlo_options.NormalSobolInnovations
        >>> path = fo.monte_carlo_options.WienerPath
        >>> payoff = fo.monte_carlo_options.PlainVanillaPayoff
        >>> opt = fo.monte_carlo_options.MonteCarloOption(50, 30, 5000,
                100, 100, 1/12, 0.1, 0.1, 0.4, inno, path, payoff)
        >>> opt.call()

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        iteration = self._sim_mc(call=True)
        if self._trace:
            self._print_trace(iteration)

        return iteration

    def put(self):
        """
        Returns an array of the average option out price per Monte Carlo Loop (mc_loop). Final option value
        is the average of the returned array.

        Returns
        -------
        numpy array

        Example
        -------
        >>> import finoptions as fo
        >>> inno = fo.monte_carlo_options.NormalSobolInnovations
        >>> path = fo.monte_carlo_options.WienerPath
        >>> payoff = fo.monte_carlo_options.PlainVanillaPayoff
        >>> opt = fo.monte_carlo_options.MonteCarloOption(50, 30, 5000,
                100, 100, 1/12, 0.1, 0.1, 0.4, inno, path, payoff)
        >>> opt.put()

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        iteration = self._sim_mc(call=False)
        if self._trace:
            self._print_trace(iteration)

        return iteration

    def _print_trace(self, iteration):
        print("\nMonte Carlo Simulation Path:\n\n")
        print("\nLoop:\t", "No\t")

        for i, _ in enumerate(iteration):
            print(
                "\nLoop:\t",
                i,
                "\t:",
                iteration[i],
                _np.sum(iteration) / (i + 1),
                end="",
            )
        print("\n")

    def _sim_mc(self, call=True):

        iteration = _np.zeros(self._mc_loops)

        # MC Iteration Loop:

        for i in range(self._mc_loops):
            #     # if ( i > 1) init = FALSE
            # Generate Innovations:
            eps = self._Innovation.sample_innovation()

            # Use Antithetic Variates if requested:
            if self._antithetic:
                eps = _np.concatenate((eps, -eps))
            #     # Standardize Variates if requested:
            if self._standardization:
                eps = (eps - _np.mean(eps)) / _np.std(eps)

                # eps = (eps-mean(eps))/sqrt(var(as.vector(eps)))

            # Calculate for each path the option price:
            path = self._Path(eps, self._sigma, self._dt, self._b)

            # so I think the original fOptions function has an error. It calcs
            # the payoff along the wrong dimensions such that it only calcs
            # along path_length number of samples vs mc_paths. I think the t()
            # is the problem.

            if call == True:
                payoff = self._Payoff(
                    path, self._S, self._K, self._t, self._r, self._b, self._sigma
                ).call()
            else:
                payoff = self._Payoff(
                    path, self._S, self._K, self._t, self._r, self._b, self._sigma
                ).put()

            tmp = _np.mean(payoff)

            if tmp == _np.inf:
                _warnings.warn(f"Warning: mc_loop {i} returned Inf.")
                return (eps, path, payoff)

            iteration[i] = tmp

        return iteration



from .base import GreeksFDM, Option
import numpy as _np
from scipy.optimize import root_scalar as _root_scalar
from scipy.optimize import root as _root
import sys as _sys
import warnings as _warnings
from .utils import docstring_from


class GBSOption(Option):
    """
    Calculate the Generalized Black-Scholes option price for either a call or put option.

    Parameters
    ----------
    S : float
        Level or index price.
    K : float
        Strike price.
    t : float
        Time-to-maturity in fractional years. i.e. 1/12 for 1 month, 1/252 for 1 business day, 1.0 for 1 year.
    r : float
        Risk-free-rate in decimal format (i.e. 0.01 for 1%).
    b : float
        Annualized cost-of-carry rate, e.g. 0.1 means 10%
    sigma : float
        Annualized volatility of the underlying asset. Optional if calculating implied volatility.
        Required otherwise. By default None.

    Notes
    -----
    Including a cost of carry term b, the model can
    used to price European Options on:

    b = r       stocks (Black and Scholes’ stock option model)
    b = r - q   stocks and stock indexes paying a continuous dividend yield q (Merton’s stock option model)
    b = 0       futures (Black’s futures option model)
    b = r - rf  currency options with foreign interst rate rf (Garman and Kohlhagen’s currency option model)

    Returns
    -------
    GBSOption object.

    Example
    -------
    >>> import finoptions as fo
    >>> opt = fo.GBSOption(10.0, 8.0, 1.0, 0.02, 0.01, 0.1)
    >>> opt.call()
    >>> opt.put()
    >>> opt.greeks(call=True)

    References
    ----------
    [1] Haug E.G., The Complete Guide to Option Pricing Formulas
    """

    __name__ = "GBSOption"
    __title__ = "Black Scholes Option"

    def __init__(
        self, S: float, K: float, t: float, r: float, b: float, sigma: float = None
    ):

        # call the init method of the Option class
        super().__init__(S, K, t, r, b, sigma)

        # additional init for d1 and d2
        # fmt: off
        if sigma is not None:
            self._d1 = (
                _np.log(self._S / self._K)
                + (self._b + self._sigma * self._sigma / 2) * self._t) / (self._sigma * _np.sqrt(self._t)
                )

            self._d2 = self._d1 - self._sigma * _np.sqrt(self._t)
        # fmt: on

        self._greeks = GreeksFDM(self)

    def _check_sigma(self, func):
        if self._sigma is None:
            raise ValueError(f"sigma not defined. Required for {func}() method")

    def call(self):
        """
        Returns the calculated price of a call option according to the
        Generalized Black-Scholes option price model.

        Returns
        -------
        float

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.GBSOption(10.0, 8.0, 1.0, 0.02, 0.01, 0.1)
        >>> opt.call()

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        self._check_sigma("call")
        # fmt: off
        result = (
                    self._S*_np.exp((self._b-self._r)*self._t)*self._CND(self._d1) 
                    - self._K*_np.exp(-self._r*self._t)*self._CND(self._d2)
                )
        # fmt: on

        return result

    def put(self):
        """
        Returns the calculated price of a put option according to the
        Generalized Black-Scholes option price model.

        Returns
        -------
        float

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.GBSOption(10.0, 8.0, 1.0, 0.02, 0.01, 0.1)
        >>> opt.put()

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        self._check_sigma("put")
        # fmt: off
        result = (
                    self._K * _np.exp(-self._r * self._t) * self._CND(-self._d2) 
                    - self._S * _np.exp((self._b - self._r) * self._t) * self._CND(-self._d1)
                )
        # fmt: on

        return result

    def delta(self, call: bool = True, method: str = "analytic"):
        """
        Method to return delta greek for either call or put options.

        Parameters
        ----------
        call : bool
            Returns delta greek for call option if True, else returns delta greek for put options. By default True.
        method : str
            'analytic' for analytic solution to delta. 'fdm' for Finite Difference Method. By default 'analytic'

        Returns
        -------
        float

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.GBSOption(10.0, 8.0, 1.0, 0.02, 0.01, 0.1)
        >>> opt.delta(call=True)

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        self._check_sigma("delta")

        if method == "analytic":
            if call == True:
                return _np.exp((self._b - self._r) * self._t) * self._CND(self._d1)
            else:
                return _np.exp((self._b - self._r) * self._t) * (
                    self._CND(self._d1) - 1
                )
        elif method == "fdm":
            return self._greeks.delta(call=call)

    def theta(self, call: bool = True, method: str = "analytic"):
        """
        Method to return theta greek for either call or put options.

        Parameters
        ----------
        call : bool
            Returns theta greek for call option if True, else returns theta greek for put options. By default True.
        method : str
            'analytic' for analytic solution to delta. 'fdm' for Finite Difference Method. By default 'analytic'

        Returns
        -------
        float

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.GBSOption(10.0, 8.0, 1.0, 0.02, 0.01, 0.1)
        >>> opt.theta(call=True)

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        self._check_sigma("theta")
        if method == "analytic":
            Theta1 = -(
                self._S
                * _np.exp((self._b - self._r) * self._t)
                * self._NDF(self._d1)
                * self._sigma
            ) / (2 * _np.sqrt(self._t))

            if call == True:
                return (
                    Theta1
                    - (self._b - self._r)
                    * self._S
                    * _np.exp((self._b - self._r) * self._t)
                    * self._CND(+self._d1)
                    - self._r
                    * self._K
                    * _np.exp(-self._r * self._t)
                    * self._CND(+self._d2)
                )
            else:
                return (
                    Theta1
                    + (self._b - self._r)
                    * self._S
                    * _np.exp((self._b - self._r) * self._t)
                    * self._CND(-self._d1)
                    + self._r
                    * self._K
                    * _np.exp(-self._r * self._t)
                    * self._CND(-self._d2)
                )
        elif method == "fdm":
            return self._greeks.theta(call=call)

    def vega(self, method: str = "analytic"):
        """
        Method to return vega greek for either call or put options.

        Parameters
        ----------
        None

        Returns
        -------
        float

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.GBSOption(10.0, 8.0, 1.0, 0.02, 0.01, 0.1)
        >>> opt.vega()

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        self._check_sigma("vega")
        # for both call and put options
        if method == "analytic":
            return (
                self._S
                * _np.exp((self._b - self._r) * self._t)
                * self._NDF(self._d1)
                * _np.sqrt(self._t)
            )
        elif method == "fdm":
            return self._greeks.vega()

    def rho(self, call: bool = True, method: str = "analytic"):
        """
        Method to return rho greek for either call or put options.

        Parameters
        ----------
        call : bool
            Returns rho greek for call option if True, else returns rho greek for put options. By default True.
        method : str
            'analytic' for analytic solution to delta. 'fdm' for Finite Difference Method. By default 'analytic'

        Returns
        -------
        float

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.GBSOption(10.0, 8.0, 1.0, 0.02, 0.01, 0.1)
        >>> opt.rho(call=True)

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        self._check_sigma("rho")
        if method == "analytic":
            if call == True:
                price = self.call()
                if self._b != 0:
                    result = (
                        self._t
                        * self._K
                        * _np.exp(-self._r * self._t)
                        * self._CND(self._d2)
                    )
                else:
                    result = -self._t * price
            else:
                price = self.put()
                if self._b != 0:
                    result = (
                        -self._t
                        * self._K
                        * _np.exp(-self._r * self._t)
                        * self._CND(-self._d2)
                    )
                else:
                    result = -self._t * price

            return result
        elif method == "fdm":
            # This only works if the cost to carry b is zero....
            if self._b == 0:
                return self._greeks.rho(call=call)
            else:
                raise ValueError(
                    "rho calculation versus finite difference method currently only works if b = 0"
                )

    def lamb(self, call: bool = True, method: str = "analytic"):
        """
        Method to return lambda greek for either call or put options.

        Parameters
        ----------
        call : bool
            Returns lambda greek for call option if True, else returns lambda greek for put options. By default True.

        Returns
        -------
        float

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.GBSOption(10.0, 8.0, 1.0, 0.02, 0.01, 0.1)
        >>> opt.lambda(call=True)

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        self._check_sigma("lamb")
        if method == "analytic":
            if call == True:
                price = self.call()
                result = (
                    _np.exp((self._b - self._r) * self._t)
                    * self._CND(self._d1)
                    * self._S
                    / price
                )

            else:
                price = self.put()
                result = (
                    _np.exp((self._b - self._r) * self._t)
                    * (self._CND(self._d1) - 1)
                    * self._S
                    / price
                )

            return result
        elif method == "fdm":
            return self._greeks.lamb(call=call)

    def gamma(self, method: str = "analytic"):
        """
        Method to return gamma greek for either call or put options.

        Parameters
        ----------
        None

        Returns
        -------
        float

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.GBSOption(10.0, 8.0, 1.0, 0.02, 0.01, 0.1)
        >>> opt.gamma(call=True)

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        self._check_sigma("gamma")
        # for both call and put options
        if method == "analytic":
            return (
                _np.exp((self._b - self._r) * self._t)
                * self._NDF(self._d1)
                / (self._S * self._sigma * _np.sqrt(self._t))
            )
        elif method == "fdm":
            return self._greeks.gamma()

    def c_of_c(self, call: bool = True):
        """
        Method to return C of C greek for either call or put options.

        Parameters
        ----------
        call : bool
            Returns C of C greek for call option if True, else returns C of C greek for put options. By default True.

        Returns
        -------
        float

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.GBSOption(10.0, 8.0, 1.0, 0.02, 0.01, 0.1)
        >>> opt.c_of_c(call=True)

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        self._check_sigma("c_of_c")
        if call == True:
            return (
                self._t
                * self._S
                * _np.exp((self._b - self._r) * self._t)
                * self._CND(self._d1)
            )
        else:
            return (
                -self._t
                * self._S
                * _np.exp((self._b - self._r) * self._t)
                * self._CND(-self._d1)
            )

    def greeks(self, call: bool = True):
        """
        Method to return greeks delta, theta, vegam rho, lambda, gamma and C of C for either
        call or put options.

        Parameters
        ----------
        call : bool
            Returns greeks for call option if True, else returns greeks for put options. By default True.

        Returns
        -------
        dictionary of greeks

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.GBSOption(10.0, 8.0, 1.0, 0.02, 0.01, 0.1)
        >>> opt.greeks(call=True)

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        self._check_sigma("greeks")

        # need to define entire greek dict as the base GreekFDM class
        # generates the dict using the FDM method
        gk = {
            "delta": self.delta(call),
            "theta": self.theta(call),
            "vega": self.vega(),
            "rho": self.rho(call),
            "lambda": self.lamb(call),
            "gamma": self.gamma(),
            "CofC": self.c_of_c(call=call),
        }

        return gk

    def volatility(
        self,
        price: float,
        call: bool = True,
        tol=_sys.float_info.epsilon,
        maxiter=10000,
        verbose=False,
    ):
        """
        Compute the implied volatility of the GBSOption.

        Parameters
        ----------
        price : float
            Current price of the option
        call : bool
            Returns implied volatility for call option if True, else returns implied volatility for put options. By default True.
        tol : float
            max tolerance to fit the price to. By default system tolerance.
        maxiter : int
            number of iterations to run to fit price.
        verbose : bool
            True to return full optimization details from root finder function. False to just return the implied volatility numbers.

        Returns
        -------
        float

        Example
        -------
        """

        sol = self._volatility(price, call, tol, maxiter, verbose)

        return sol


class BlackScholesOption(GBSOption):
    __name__ = "BlackScholesOption"
    __title__ = "Black Scholes Option"


class Black76Option(GBSOption):
    """
    The Black76Option pricing formula is applicable for valuing European call
    and European put options on commodity futures. The exact nature of the
    underlying commodity varies and may be anything from a precious metal such
    as gold or silver to agricultural products.

    Parameters
    ----------
    FT : float
        Futures price.
    K : float
        Strike price.
    t : float
        Time-to-maturity in fractional years. i.e. 1/12 for 1 month, 1/252 for 1 business day, 1.0 for 1 year.
    r : float
        Risk-free-rate in decimal format (i.e. 0.01 for 1%).
    sigma : float
        Annualized volatility of the underlying asset. Optional if calculating implied volatility.
        Required otherwise. By default None.
    """

    __name__ = "Black76Option"
    __title__ = "Black 1977 Option"

    def __init__(self, FT, K, t, r, sigma=None):
        super().__init__(FT, K, t, r, b=0, sigma=sigma)
        # self._FT = self._S

    def get_params(self):
        return {
            "FT": self._S,
            "K": self._K,
            "t": self._t,
            "r": self._r,
            "b": self._b,
            "sigma": self._sigma,
        }

    def set_param(self, x: str, value: float):
        """
        Method to change a parameter once the class hass been initiatized.

        Parameters
        ----------
        x : str
            Name of parameter to change
        value : float
            New value to give to parameter

        Returns
        -------
        None
        """

        tmp = self.get_params()
        tmp[x] = value

        del tmp["b"]  # delete b prior to re-init

        self.__init__(**tmp)


class MiltersenSchwartzOption(Option):
    """
    The MiltersenSchwartzOption class allows for pricing options on commodity futures. The model is a three
    factor model with stochastic futures prices, term structures of convenience yields and interest rates.
    The model is based on lognormal distributed commodity prices and normal distributed continuously compounded
    forward interest rates and futures convenience yields.

    The Miltersen Schwartz Option model is a three factor model with stochastic futures prices,term structures
    and convenience yields, and interest rates. The model is based on log-normal distributed commodity prices
    and normal distributed continuously compounded forward interest rates and future convenience yields.

    Parameters
    ----------
    Pt : float
        The zero coupon bond that expires on the option maturity.
    FT : float
        The futures price.
    K : float
        The strike price.
    t : float
        Time-to-maturity in fractional years. i.e. 1/12 for 1 month, 1/252 for 1 business day, 1.0 for 1 year.
    T : float
        Time-to-maturity in fractional years. i.e. 1/12 for 1 month, 1/252 for 1 business day, 1.0 for 1 year.
    sigmaS : float
        The annualized volatility of the spot commodity price (S), e.g.  0.25 means 25%
    sigmaE : float
        The annualized volatility of the the future convenience yield (E), e.g. 0.25 means 25%
    sigmaF : float
        The annualized volatility of the the forward interest rate (F), e.g.  0.25 means 25%
    rhoSE : float
        The correlations between the spot commodity price and the future convenience yield (SE)
    rhoSF : float
        The correlations between the spot commodity price and the forward interest rate (SF)
    rhoEF : float
        The correlations between the forward interest rate and the future convenience yield (EF)
    KappaE : float
        The speed of mean reversion of the forward interest rate (E)
    KappaF : float
        The speed of mean reversion of the convenience yield (F)

    References
    ----------
    [1] Haug E.G., The Complete Guide to Option Pricing Formulas
    [2] Miltersen  K.,  Schwartz  E.S.  (1998);Pricing  of  Options  on  Commodity  Futures  with  Stochastic Term Structuures of Convenience Yields and Interest Rates, Journal of Financial and Quantitative Analysis 33, 33–59
    """

    __name__ = "MiltersenSchwartzOption"
    __title__ = "Miltersen Schwartz Option"

    def __init__(
        self,
        Pt: float,
        FT: float,
        K: float,
        t: float,
        T: float,
        sigmaS: float,
        sigmaE: float,
        sigmaF: float,
        rhoSE: float,
        rhoSF: float,
        rhoEF: float,
        KappaE: float,
        KappaF: float,
    ):
        self._Pt = Pt
        self._FT = FT
        self._K = K
        self._t = t
        self._T = T
        self._sigmaS = sigmaS
        self._sigmaE = sigmaE
        self._sigmaF = sigmaF
        self._rhoSE = rhoSE
        self._rhoSF = rhoSF
        self._rhoEF = rhoEF
        self._KappaE = KappaE
        self._KappaF = KappaF

        # fmt: off
        self._vz = (
            self._sigmaS**2*self._t+2*self._sigmaS*(self._sigmaF*self._rhoSF*1/self._KappaF*(self._t-1/self._KappaF*
            _np.exp(-self._KappaF*self._T)*(_np.exp(self._KappaF*self._t)-1))-self._sigmaE*self._rhoSE*1/self._KappaE*
            (self._t-1/self._KappaE*_np.exp(-self._KappaE*self._T)*(_np.exp(self._KappaE*self._t)-1)))+self._sigmaE**2*
            1/self._KappaE**2*(self._t+1/(2*self._KappaE)*_np.exp(-2*self._KappaE*self._T)*(_np.exp(2*self._KappaE*self._t)-
            1)-2*1/self._KappaE*_np.exp(-self._KappaE*self._T)*(_np.exp(self._KappaE*self._t)-1))+self._sigmaF**2*
            1/self._KappaF**2*(self._t+1/(2*self._KappaF)*_np.exp(-2*self._KappaF*self._T)*(_np.exp(2*self._KappaF*self._t)-
            1)-2*1/self._KappaF*_np.exp(-self._KappaF*self._T)*(_np.exp(self._KappaF*self._t)-1))-2*self._sigmaE*
            self._sigmaF*self._rhoEF*1/self._KappaE*1/self._KappaF*(self._t-1/self._KappaE*_np.exp(-self._KappaE*self._T)*
            (_np.exp(self._KappaE*self._t)-1)-1/self._KappaF*_np.exp(-self._KappaF*self._T)*(_np.exp(self._KappaF*self._t)-
            1)+1/(self._KappaE+self._KappaF)*_np.exp(-(self._KappaE+self._KappaF)*self._T)*(_np.exp((self._KappaE+self._KappaF)*
            self._t)-1))
        )
        self._vxz = (
            self._sigmaF*1/self._KappaF*(self._sigmaS*self._rhoSF*(self._t-1/self._KappaF*(1-_np.exp(-self._KappaF*
            self._t)))+self._sigmaF*1/self._KappaF*(self._t-1/self._KappaF*_np.exp(-self._KappaF*self._T)*(_np.exp(self._KappaF*
            self._t)-1)-1/self._KappaF*(1-_np.exp(-self._KappaF*self._t))+1/(2*self._KappaF)*_np.exp(-self._KappaF*
            self._T)*(_np.exp(self._KappaF*self._t)-_np.exp(-self._KappaF*self._t)))-self._sigmaE*self._rhoEF*1/self._KappaE*
            (self._t-1/self._KappaE*_np.exp(-self._KappaE*self._T)*(_np.exp(self._KappaE*self._t)-1)-1/self._KappaF*(1-
            _np.exp(-self._KappaF*self._t))+1/(self._KappaE+self._KappaF)*_np.exp(-self._KappaE*self._T)*
            (_np.exp(self._KappaE*self._t)-_np.exp(-self._KappaF*self._t))))
        )
        
        self._vz = _np.sqrt(self._vz)

        self._d1 = (_np.log(self._FT/self._K)-self._vxz+self._vz**2/2)/self._vz
        self._d2 = (_np.log(self._FT/self._K)-self._vxz-self._vz**2/2)/self._vz

        # fmt: on

        self._greeks = GreeksFDM(self)

    def call(self):
        """
        Returns the calculated price of a call option according to the
        Miltersen Schwartz Option option price model.

        Returns
        -------
        float

        Example
        -------
        >>> import finoptions as fo
        >>> import numpy as np
        >>> opt = fo.MiltersenSchwartzOption(Pt=np.exp(-0.05/4), FT=95, K=80, t=1/4, T=1/2, sigmaS=0.2660,
                    sigmaE=0.2490, sigmaF=0.0096, rhoSE=0.805, rhoSF=0.0805, rhoEF=0.1243, KappaE=1.045, KappaF=0.200)
        >>> opt.call()

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """

        # fmt: off
        result = self._Pt*(self._FT*_np.exp(-self._vxz)*self._CND(self._d1)-self._K*self._CND(self._d2))
        # fmt: on

        return result

    def put(self):
        """
        Returns the calculated price of a call option according to the
        Miltersen Schwartz Option option price model.

        Returns
        -------
        float

        Example
        -------
        >>> import finoptions as fo
        >>> import numpy as np
        >>> opt = fo.MiltersenSchwartzOption(Pt=np.exp(-0.05/4), FT=95, K=80, t=1/4, T=1/2, sigmaS=0.2660,
                    sigmaE=0.2490, sigmaF=0.0096, rhoSE=0.805, rhoSF=0.0805, rhoEF=0.1243, KappaE=1.045, KappaF=0.200)
        >>> opt.put()

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """

        # fmt: off
        result = self._Pt*(self._K*self._CND(-self._d2)-self._FT*_np.exp(-self._vxz)*self._CND(-self._d1))
        # fmt: on

        return result

    def get_params(self):
        return {
            "Pt": self._Pt,
            "FT": self._FT,
            "K": self._K,
            "t": self._t,
            "T": self._T,
            "sigmaS": self._sigmaS,
            "sigmaE": self._sigmaE,
            "sigmaF": self._sigmaF,
            "rhoSE": self._rhoSE,
            "rhoSF": self._rhoSF,
            "rhoEF": self._rhoEF,
            "KappaE": self._KappaE,
            "KappaF": self._KappaF,
        }

    def delta(self, call: bool = True):
        """
        Method to return delta greek for either call or put options using Finite Difference Methods.

        Parameters
        ----------
        call : bool
            Returns delta greek for call option if True, else returns delta greek for put options. By default True.

        Returns
        -------
        float
        """
        fd = self._greeks._make_partial_der("FT", call, self, n=1)

        return fd(self._FT) * 1

    @docstring_from(GreeksFDM.theta)
    def theta(self, call: bool = True):
        return self._greeks.theta(call=call)

    def vega(self):
        """
        Method to return vega greek for either call or put options using Finite Difference Methods.

        Parameters
        ----------
        None

        Returns
        -------
        float
        """
        # same for both call and put options
        fdS = self._greeks._make_partial_der("sigmaS", True, self, n=1)
        fdE = self._greeks._make_partial_der("sigmaE", True, self, n=1)
        fdF = self._greeks._make_partial_der("sigmaF", True, self, n=1)
        result = _np.array(
            (
                fdS(self._sigmaS) * 1,
                fdE(self._sigmaE) * 1,
                fdF(self._sigmaF) * 1,
            )
        )

        # if returning a multi-dim array because arrays were passed as inputs, transpose prior
        # to returing results so that each element of the array input corresponds to a row
        # in the result
        if self._check_array(fdS(self._sigmaS) * 1) == True:
            result = result.T

        return result

    def greeks(self, call: bool = True):
        gk = {
            "delta": self.delta(call),
            "theta": self.theta(call),
            "vega": self.vega(),
            "lambda": self.lamb(call),
            "gamma": self.gamma(),
            # "CofC": self.c_of_c(call),
        }

        return gk

    def lamb(self, call: bool = True):
        """
        Method to return lambda greek for either call or put options.

        Parameters
        ----------
        call : bool
            Returns lambda greek for call option if True, else returns lambda greek for put options. By default True.

        Returns
        -------
        float
        """
        if call == True:
            price = self.call()
        else:
            price = self.put()
        return self.delta(call=call) * self._FT / price

    def gamma(self):
        """
        Method to return gamma greek for either call or put options.

        Parameters
        ----------
        None

        Returns
        -------
        float
        """
        # same for both call and put options
        fd = self._greeks._make_partial_der("FT", True, self, n=2)
        return fd(self._FT) * 1



from abc import ABC, abstractmethod
import numpy as _np
from math import atan as _atan
import copy as _copy
import numdifftools as _nd
import sys as _sys
import warnings as _warnings
from scipy.optimize import root_scalar as _root_scalar
from scipy.optimize import root as _root
import scipy.optimize as _opt


class _Base(ABC):
    def __init__(self):
        pass

    def _NDF(self, x):
        """
        Calculate the normal distribution function of x

        Parameters
        ----------

        x : float
            Value to calculate the normal distribution function for x.

        Returns
        -------
        float

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """

        return _np.exp(-x * x / 2) / _np.sqrt(8 * _atan(1))

    def _CND(self, x):
        """
        Calculate the cumulated normal distribution function of x

        Parameters
        ----------

        x : float
            Value to calculate the cumulated normal distribution function for x.

        Returns
        -------
        float

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """

        k = 1 / (1 + 0.2316419 * abs(x))
        a1 = 0.319381530
        a2 = -0.356563782
        a3 = 1.781477937
        a4 = -1.821255978
        a5 = 1.330274429
        result = (
            self._NDF(x)
            * (a1 * k + a2 * k ** 2 + a3 * k ** 3 + a4 * k ** 4 + a5 * k ** 5)
            - 0.5
        )
        result = 0.5 - result * _np.sign(x)

        return result

    def _CBND(self, x1, x2, rho):
        """
        Calculate the cumulative bivariate normal distribution function.

        Haug E.G., The Complete Guide to Option Pricing Formulas

        Compute:
        Take care for the limit rho = +/- 1
        """

        a = x1
        b = x2
        if abs(rho) == 1:
            rho = rho - (1e-12) * _np.sign(rho)
        # cat("\n a - b - rho :"); print(c(a,b,rho))
        X = [0.24840615, 0.39233107, 0.21141819, 0.03324666, 0.00082485334]
        y = [0.10024215, 0.48281397, 1.0609498, 1.7797294, 2.6697604]
        a1 = a / _np.sqrt(2 * (1 - rho ** 2))
        b1 = b / _np.sqrt(2 * (1 - rho ** 2))
        if (a <= 0) & (b <= 0) & (rho <= 0):
            Sum1 = 0
            for I in range(0, 5):
                for j in range(0, 5):
                    Sum1 = Sum1 + X[I] * X[j] * _np.exp(
                        a1 * (2 * y[I] - a1)
                        + b1 * (2 * y[j] - b1)
                        + 2 * rho * (y[I] - a1) * (y[j] - b1)
                    )
            result = _np.sqrt(1 - rho ** 2) / _np.pi * Sum1
            return result

        if (a <= 0) & (b >= 0) & (rho >= 0):
            result = self._CND(a) - self._CBND(a, -b, -rho)
            return result

        if (a >= 0) & (b <= 0) & (rho >= 0):
            result = self._CND(b) - self._CBND(-a, b, -rho)
            return result

        if (a >= 0) & (b >= 0) & (rho <= 0):
            result = self._CND(a) + self._CND(b) - 1 + self._CBND(-a, -b, rho)
            return result

        if (a * b * rho) >= 0:
            rho1 = (
                (rho * a - b)
                * _np.sign(a)
                / _np.sqrt(a ** 2 - 2 * rho * a * b + b ** 2)
            )
            rho2 = (
                (rho * b - a)
                * _np.sign(b)
                / _np.sqrt(a ** 2 - 2 * rho * a * b + b ** 2)
            )
            delta = (1 - _np.sign(a) * _np.sign(b)) / 4
            result = self._CBND(a, 0, rho1) + self._CBND(b, 0, rho2) - delta
            return result


class Derivative(_Base):
    pass

    @abstractmethod
    def get_params(self):
        pass

    def copy(self):
        return _copy.deepcopy(self)


class GreeksFDM:
    """
    Greek calculation class for composition of Option classes

    Parameters
    ----------
    opt : Option class
        Option class with Option values to calculate greeks from

    """

    def __init__(self, opt):

        if True:  # issubclass(opt, Option):
            self._opt = opt
        else:
            raise ValueError(
                f"Parameter opt is not of the Option class, it is of the {type(opt)} class"
            )

    def _make_partial_der(self, wrt, call, opt, **kwargs):
        """
        Create monad from Option methods call and put for use
        in calculating the partial derivatives or greeks with
        respect to wrt.
        """

        def _func(x):
            tmp = opt.copy()
            tmp.set_param(wrt, x)
            if call == True:
                return tmp.call()
            else:
                return tmp.put()

        fd = _nd.Derivative(_func, **kwargs)

        return fd

    def delta(self, call: bool = True):
        """
        Method to return delta greek for either call or put options using Finite Difference Methods.

        Parameters
        ----------
        call : bool
            Returns delta greek for call option if True, else returns delta greek for put options. By default True.

        Returns
        -------
        float
        """
        fd = self._make_partial_der("S", call, self._opt, n=1)

        # multiple by 1 to return float vs array for single element arrays. Multi-element arrays returned as normal
        return fd(self._opt._S) * 1

    def theta(self, call: bool = True):
        """
        Method to return theta greek for either call or put options using Finite Difference Methods.

        Parameters
        ----------
        call : bool
            Returns theta greek for call option if True, else returns theta greek for put options. By default True.

        Returns
        -------
        float
        """
        fd = self._make_partial_der("t", call, self._opt, n=1, step=1 / 252)

        # multiple by 1 to return float vs array for single element arrays. Multi-element arrays returned as normal
        return fd(self._opt._t) * -1

    def vega(self):
        """
        Method to return vega greek for either call or put options using Finite Difference Methods.

        Parameters
        ----------
        None

        Returns
        -------
        float
        """
        # same for both call and put options
        fd = self._make_partial_der("sigma", True, self._opt, n=1)

        # multiple by 1 to return float vs array for single element arrays. Multi-element arrays returned as normal
        return fd(self._opt._sigma) * 1

    def rho(self, call: bool = True):
        """
        Method to return rho greek for either call or put options using Finite Difference Methods.

        Parameters
        ----------
        call : bool
            Returns rho greek for call option if True, else returns rho greek for put options. By default True.

        Returns
        -------
        float
        """
        # This only works if the cost to carry b is zero....
        fd = self._make_partial_der("r", call, self._opt, n=1)

        # multiple by 1 to return float vs array for single element arrays. Multi-element arrays returned as normal
        return fd(self._opt._r) * 1

    def lamb(self, call: bool = True):
        """
        Method to return lambda greek for either call or put options using Finite Difference Methods.

        Parameters
        ----------
        call : bool
            Returns lambda greek for call option if True, else returns lambda greek for put options. By default True.

        Returns
        -------
        float
        """
        if call == True:
            price = self._opt.call()
        else:
            price = self._opt.put()
        return self.delta(call=call) * self._opt._S / price

    def gamma(self):
        """
        Method to return gamma greek for either call or put options using Finite Difference Methods.

        Parameters
        ----------
        None

        Returns
        -------
        float
        """
        # same for both call and put options
        fd = self._make_partial_der("S", True, self._opt, n=2)

        # multiple by 1 to return float vs array for single element arrays. Multi-element arrays returned as normal
        return fd(self._opt._S) * 1

    def greeks(self, call: bool = True):
        """
        Method to return greeks as a dictiontary for either call or put options using Finite Difference Methods.

        Parameters
        ----------
        None

        Returns
        -------
        float
        """
        gk = {
            "delta": self.delta(call),
            "theta": self.theta(call),
            "vega": self.vega(),
            "rho": self.rho(call),
            "lambda": self.lamb(call),
            "gamma": self.gamma(),
        }

        return gk


class Option(Derivative):
    """
    Base class for options

    Parameters
    ----------
    S : float
        Level or index price.
    K : float
        Strike price.
    t : float
        Time-to-maturity in fractional years. i.e. 1/12 for 1 month, 1/252 for 1 business day, 1.0 for 1 year.
    r : float
        Risk-free-rate in decimal format (i.e. 0.01 for 1%).
    b : float
        Annualized cost-of-carry rate, e.g. 0.1 means 10%
    sigma : float
        Annualized volatility of the underlying asset. Optional if calculating implied volatility.
        Required otherwise. By default None.

    Note
    ----
    that setting:
    b = r we get Black and Scholes’ stock option model
    b = r-q we get Merton’s stock option model with continuous dividend yield q
    b = 0 we get Black’s futures option model
    b = r-rf we get Garman and Kohlhagen’s currency option model with foreign
    interest rate rf
    """

    @property
    @abstractmethod
    def __name__(self):
        pass

    @property
    @abstractmethod
    def __title__(self):
        pass

    def __init__(self, S: float, K: float, t: float, r: float, b: float, sigma: float):
        self._S = S
        self._K = K
        self._t = t
        self._r = r
        self._b = b
        self._sigma = sigma

    def get_params(self):
        return {
            "S": self._S,
            "K": self._K,
            "t": self._t,
            "r": self._r,
            "b": self._b,
            "sigma": self._sigma,
        }

    def set_param(self, x: str, value: float):
        """
        Method to change a parameter once the class hass been initiatized.

        Parameters
        ----------
        x : str
            Name of parameter to change
        value : float
            New value to give to parameter

        Returns
        -------
        None
        """

        tmp = self.get_params()
        tmp[x] = value

        self.__init__(**tmp)

    # @abstractmethod
    # def delta(self):
    #     pass

    # @abstractmethod
    # def theta(self):
    #     pass

    # @abstractmethod
    # def vega(self):
    #     pass

    # @abstractmethod
    # def lamb(self):
    #     pass

    # @abstractmethod
    # def gamma(self):
    #     pass

    # @abstractmethod
    # def greeks(self):
    #     pass

    def summary(self, printer=True):
        """
        Print summary report of option

        Parameters
        ----------
        printer : bool
            True to print summary. False to return a string.
        """
        out = f"Title: {self.__title__} Valuation\n\nParameters:\n\n"

        params = self.get_params()

        for p in params:
            out += f"  {p} = {self._check_string(params[p])}\n"

        try:
            # if self._sigma or its variations are not None add call and put prices
            if isinstance(self.call(), _np.ndarray):
                c = self._check_string(self.call().round(2))
                p = self._check_string(self.put().round(2))
                price = f"\nOption Price:\n\n  call: {c}\n  put: {p}"
            else:
                price = f"\nOption Price:\n\n  call: {round(self.call(),6)}\n  put: {round(self.put(),6)}"
            out += price
        except:
            pass

        if printer == True:
            print(out)
        else:
            return out

    def __str__(self):
        return self.summary(printer=False)

    def __repr__(self):
        out = f"{self.__name__}("
        params = self.get_params()

        for p in params:
            out = out + str(self._check_string(params[p])) + ", "

        # get rid of trailing comman and close pararenthesis
        out = out[:-2]
        out += ")"

        return out

    def _check_string(self, x):
        """
        helper function for summary method. Checks to see if the variable x is a numpy
        ndarray type and if it's length is greater than 6. If so, shortens the repsentation
        so that it fits.
        """
        if isinstance(x, _np.ndarray):
            if x.shape[0] > 6:
                return _np.array2string(x.round(2), threshold=6)

        return x

    def _check_array(self, *args):
        """
        helper function to return True if any args past are numpy ndarrays
        """
        for a in args:
            # allows size 1 arrays. somethings greeks return size 1 arrays
            if isinstance(a, _np.ndarray) and a.size > 1:
                return True

        return False

    def _max_array(self, *args):
        """
        helper function to get largest ndarray. Assumes at
        least one array.
        """
        maxArray = _np.array([0])
        for a in args:
            if isinstance(a, _np.ndarray):
                if maxArray.size < a.size:
                    maxArray = a

        return maxArray

    def _func(sigma, obj, call, price):
        """
        helper function to be used for volatility root finding.
        """
        temp = obj.copy()
        temp.set_param("sigma", sigma)
        if call == True:
            return price - temp.call()
        else:
            return price - temp.put()

    def _volatility(
        self,
        price: float,
        call: bool = True,
        tol=_sys.float_info.epsilon,
        maxiter=10000,
        verbose=False,
        _func=_func,
    ):
        """
        Compute the implied volatility of the GBSOption.

        Parameters
        ----------
        price : float
            Current price of the option
        call : bool
            Returns implied volatility for call option if True, else returns implied volatility for put options. By default True.
        tol : float
            max tolerance to fit the price to. By default system tolerance.
        maxiter : int
            number of iterations to run to fit price.
        verbose : bool
            True to return full optimization details from root finder function. False to just return the implied volatility numbers.

        Returns
        -------
        float

        Example
        -------
        """
        if self._sigma is not None:
            _warnings.warn("sigma is not None but calculating implied volatility.")

        # check to see if arrays were past vs scalars.
        if self._check_array(price, *self.get_params().values()):
            # if arrays, use root function
            a = self._max_array(price, *self.get_params().values())
            if verbose == True:
                sol = _root(_func, args=(self, call, price), x0=_np.ones_like(a))
            else:
                sol = _root(_func, args=(self, call, price), x0=_np.ones_like(a)).x
        else:
            # if scalars use root_scalar function
            if verbose == True:
                sol = _root_scalar(
                    _func,
                    (self, call, price),
                    bracket=[-10, 10],
                    xtol=tol,
                    maxiter=maxiter,
                )
                # sol = _opt.brentq(_func, 1, 10000, xtol=tol, maxiter=maxiter)
            else:
                sol = _root_scalar(
                    _func,
                    (self, call, price),
                    bracket=[-10, 10],
                    xtol=tol,
                    maxiter=maxiter,
                ).root

        return sol



from abc import ABC, abstractmethod
from scipy.stats import qmc as _qmc, norm as _norm
import numpy as _np


class Innovations(ABC):
    def __init__(self, mc_paths, path_length, eps=None):
        self.mc_paths = mc_paths
        self.path_length = path_length
        self._eps = eps  # for testing only

    @abstractmethod
    def sample_innovation(self, init=False):
        pass


class NormalSobolInnovations(Innovations):
    """
    Normal Sobol Innovation Class for use with Monte Carlo Options Class

    Parameters
    ----------
    mc_paths : int
        Number of monte carlo samples per loop - total number of samples is mc_paths * mc_loops
    path_length : int
        Path length should be a power of 2 (i.e. 2**m) to generate stable paths. See
        statsmodels.stats.qmc.Sobol for more details.

    Returns
    -------
    Innovation Class
    """

    def sample_innovation(self, scramble=True):
        sobol = self._get_sobol(scramble)

        # avoid inf
        while _np.abs(_np.max(sobol)) == _np.inf:
            sobol = self._get_sobol(scramble)

        if self._eps is None:
            return sobol
        else:
            # for testing only since fOptions sobol innovations
            # return different data from statsmodels
            return self._eps

    def _get_sobol(self, scramble):
        sobol = _qmc.Sobol(self.path_length, scramble=scramble).random(self.mc_paths)
        if scramble == False:
            # add new sample since if not scrambled first row is zero which leads to -inf when normalized
            sobol = sobol[1:]
            sobol = _np.append(
                sobol,
                _qmc.Sobol(self.path_length, scramble=scramble)
                .fast_forward(self.mc_paths)
                .random(1),
                axis=0,
            )
        sobol = _norm.ppf(sobol)

        return sobol



from ..base import GreeksFDM, Option as _Option
from ..vanillaoptions import GBSOption as _GBSOption
import numpy as _np
from scipy.optimize import root_scalar as _root_scalar
import sys as _sys
import warnings as _warnings
import numdifftools as _nd
from ..utils import docstring_from
from matplotlib import pyplot as _plt


class BiTreePlotter:
    def __init__(self, tree):
        self._tree = tree

    def plot(self, dx=-0.025, dy=0.4, size=12, digits=2, **kwargs):
        depth = self._tree.shape[1] - 1

        fig = _plt.figure(**kwargs)
        for i in range(depth):
            x = [1, 0, 1]
            for j in range(i):
                x.append(0)
                x.append(1)
            x = _np.array(x) + i
            y = _np.arange(-(i + 1), i + 2)[::-1]
            _plt.plot(x, y, "bo-")

            pts = list(zip(x, y))

            for p in pts:
                value = round(self._tree[self._convert_to_index(p)][p[0]], digits)
                _plt.annotate(value, xy=p, xytext=(p[0] + dx, p[1] + dy), fontsize=size)

        _plt.xlabel("n", fontsize=size)
        _plt.ylabel("Option Value", fontsize=size)
        _plt.title("Option Tree", fontsize=int(size * 1.3))
        return fig

    def _convert_to_index(self, p):
        # converts tree y values into indices for the tree matrix
        step = -2
        max = p[0]
        if max > 0:
            a = _np.arange(max, -max + step, step)
            return int(_np.where(a == p[1])[0])
        else:
            return 0


class TriTreePlotter:
    def __init__(self, tree):
        self._tree = tree

    def plot(self, dx=-0.025, dy=0.4, size=12, digits=2, **kwargs):
        n = self._tree.shape[1] - 1
        fig = _plt.figure(**kwargs)
        for i in range(n + 1):
            for j in range(-i + 1, i):
                self._plot_node(i - 1, j, dx, dy, size, digits)

        _plt.annotate(
            round(self._tree[n, 0], digits),
            xy=(0, 0),
            xytext=(+dx, +dy),
            fontsize=size,
        )
        _plt.xlabel("n", fontsize=size)
        _plt.ylabel("Option Value", fontsize=size)
        _plt.title("Option Tree", fontsize=int(size * 1.3))
        return fig

    def _plot_node(self, i, j, dx, dy, size, digits):
        for c in [-1, 0, 1]:
            depth = self._tree.shape[1] - 1
            x = _np.array([0, 1]) + i
            y = _np.array([0, c]) + j
            _plt.plot(x, y, "bo-")
            p = (x[1], y[1])
            # print(p)
            # print(p[0], p[1] + 2)
            _plt.annotate(
                round(self._tree[-y[1] + depth][x[1]], digits),
                xy=p,
                xytext=(p[0] + dx, p[1] + dy),
                fontsize=size,
            )


class CRRBinomialTreeOption(_Option):
    """
    Binomial models were first suggested by Cox, Ross and Rubinstein (1979), CRR,
    and then became widely used because of its intuition and easy implementation. Binomial trees are
    constructed on a discrete-time lattice. With the time between two trading events shrinking to zero,
    the evolution of the price converges weakly to a lognormal diffusion. Within this mode the European
    options value converges to the value given by the Black-Scholes formula.

    Notes
    -----
    The model described here is a version of the CRR Binomial
    Tree model. Including a cost of carry term b, the model can
    used to price European and American Options on:

    b = r       stocks
    b = r - q   stocks and stock indexes paying a continuous dividend yield q
    b = 0       futures
    b = r - rf  currency options with foreign interst rate rf

    Parameters
    ----------
    S : float
        Level or index price.
    K : float
        Strike price.
    t : float
        Time-to-maturity in fractional years. i.e. 1/12 for 1 month, 1/252 for 1 business day, 1.0 for 1 year.
    r : float
        Risk-free-rate in decimal format (i.e. 0.01 for 1%).
    b : float
        Annualized cost-of-carry rate, e.g. 0.1 means 10%
    sigma : float
        Annualized volatility of the underlying asset. Optional if calculating implied volatility.
        Required otherwise. By default None.
    type : str
        "european" to price European options, "american" to price American options. By default "european"
    n : int
        Number of time steps to use. By default 5.

    Returns
    -------
    CRRBinomialTreeOption object.

    Example
    -------
    >>> import finoptions as fo
    >>> opt = fo.binomial_tree_options.CRRBinomialTreeOption(S=50, K=50, t=5/12, r=0.1, b=0., sigma=0.4)
    >>> opt.call()
    >>> opt.put()
    >>> opt.greeks(call=True)

    References
    ----------
    [1] Haug E.G., The Complete Guide to Option Pricing Formulas
    """

    __name__ = "CRRBinomialTreeOption"
    __title__ = "CRR Binomial Tree Model"

    def __init__(
        self,
        S: float,
        K: float,
        t: float,
        r: float,
        b: float,
        sigma: float,
        type: str = "european",
        n: int = 5,
    ):
        self._S = S
        self._K = K
        self._t = t
        self._r = r
        self._b = b
        self._sigma = sigma
        self._n = n
        self._type = type

        self._greeks = GreeksFDM(self)

    def get_params(self):
        return {
            "S": self._S,
            "K": self._K,
            "t": self._t,
            "r": self._r,
            "b": self._b,
            "sigma": self._sigma,
            "type": self._type,
            "n": self._n,
        }

    def call(self, tree: bool = False):
        """
        Returns the calculated price of a call option according to the
        Cox-Ross-Rubinstein Binomial Tree option price model.

        Parameters
        ----------
        tree : bool
            If True, returns the bionomial option as a tree-matrix to show the evolution of the option
            value.

        Returns
        -------
        float or tree-matrix

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.binomial_tree_options.CRRBinomialTreeOption(S=50, K=50, t=5/12, r=0.1, b=0.1, sigma=0.4, n=5, type='european')
        >>> opt.call()

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """

        z = 1
        return self._calc_price(z, self._n, self._type, tree)

    def put(self, tree: bool = False):
        """
        Returns the calculated price of a put option according to the
        Cox-Ross-Rubinstein Binomial Tree option price model.

        Parameters
        ----------
        tree : bool
            If True, returns the bionomial option as a tree-matrix to show the evolution of the option
            value.

        Returns
        -------
        float or tree-matrix

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.binomial_tree_options.CRRBinomialTreeOption(S=50, K=50, t=5/12, r=0.1, b=0.1, sigma=0.4, n=5, type='european')
        >>> opt.put()

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """

        z = -1
        return self._calc_price(z, self._n, self._type, tree)

    def summary(self, printer=True):
        """
        Print summary report of option

        Parameters
        ----------
        printer : bool
            True to print summary. False to return a string.
        """
        out = f"Title: {self.__title__} Valuation\n\nParameters:\n\n"

        params = self.get_params()

        for p in params:
            out += f"  {p} = {self._check_string(params[p])}\n"

        try:
            # if self._sigma or its variations are not None add call and put prices
            if isinstance(self.call(), _np.ndarray):
                c = self._check_string(self.call().round(2))
                p = self._check_string(self.put().round(2))
                price = f"\nOption Price:\n\n  call-{self._type}: {c}\n  put-{self._type}: {p}"
            else:
                price = f"\nOption Price:\n\n  call-{self._type}: {round(self.call(),6)}\n  put-{self._type}: {round(self.put(),6)}"
            out += price
        except:
            pass

        if printer == True:
            print(out)
        else:
            return out

    @docstring_from(GreeksFDM.delta)
    def delta(self, call: bool = True):
        return self._greeks.delta(call=call)

    @docstring_from(GreeksFDM.theta)
    def theta(self, call: bool = True):
        return self._greeks.theta(call=call)

    @docstring_from(GreeksFDM.vega)
    def vega(self):
        return self._greeks.vega()

    @docstring_from(GreeksFDM.rho)
    def rho(self, call: bool = True):
        return self._greeks.rho(call=call)

    @docstring_from(GreeksFDM.lamb)
    def lamb(self, call: bool = True):
        return self._greeks.lamb(call=call)

    @docstring_from(GreeksFDM.gamma)
    def gamma(self):
        return self._greeks.gamma()

    @docstring_from(GreeksFDM.greeks)
    def greeks(self, call: bool = True):
        return self._greeks.greeks(call=call)

    # fmt: off
    def _calc_price(self, z, n, type, tree):
        dt = self._t / n
        u = _np.exp(self._sigma * _np.sqrt(dt))
        d = 1 / u
        p = (_np.exp(self._b * dt) - d) / (u - d)
        Df = _np.exp(-self._r * dt)

        OptionValue = z * (self._S * u ** _np.arange(0, n + 1) * d ** _np.arange(n, -1, -1) - self._K)
        OptionValue = (_np.abs(OptionValue) + OptionValue) / 2 # takes the max vs zero

        if type == "european":
            out = self._euro(OptionValue, n, Df, p, tree)
        elif type == "american":
            out = self._amer(OptionValue, n, Df, p, self._K, d, self._S, u, z, tree)

        if tree == False:
            return out[0]
        else:
            return out

    def _euro(self, OptionValue, n, Df, p, tree=False):
        tr = OptionValue.copy()
        for j in _np.arange(0, n)[::-1]:
            tr = _np.append(tr, _np.zeros(n-j))
            for i in _np.arange(0, j + 1):
                OptionValue[i] = (p * OptionValue[i + 1] + (1 - p) * OptionValue[i]) * Df
                tr = _np.append(tr, OptionValue[i])

        if tree == True:
            tr = _np.reshape(tr[::-1],(n+1,n+1)).T
            return tr
        else:
            return OptionValue

    def _amer(self, OptionValue, n, Df, p, K, d, S, u, z, tree=False):
        tr = OptionValue.copy()
        for j in _np.arange(0, n)[::-1]:
            tr = _np.append(tr, _np.zeros(n-j))
            for i in _np.arange(0, j + 1):
                OptionValue[i] = max(
                    (z * (S * u ** i * d ** (abs(i - j)) - K)),
                    (p * OptionValue[i + 1] + (1 - p) * OptionValue[i]) * Df,
                )
                tr = _np.append(tr, OptionValue[i])
        if tree == True:
            tr = _np.reshape(tr[::-1],(n+1,n+1)).T
            return tr
        else:
            return OptionValue
    # fmt: on

    def plot(self, call=True, dx=-0.025, dy=0.4, size=12, digits=2, **kwargs):
        """
        Method to plot the binomial tree values

        Parameters
        ----------
        call : bool
            Set to True to plot a call option. Otherwise set to False for a put option. By default True.
        dx : float
            x-offset for the node values in the same units at the call or put option value.
        dy : float
            y-offset for the node values in the same units at the call or put option value.
        size : float
            Font size for node and axis labels. By default 12. Title is 30% larger.
        digits : int
            Number of significant figures to show for option values. By default 2.
        **kwargs
            paramters to pass to matplotlib

        Returns
        -------
        matplotlib figure object

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.binomial_tree_options.CRRBinomialTreeOption(S=50, K=50, t=5/12, r=0.1, b=0.1, sigma=0.4, n=5, type='american')
        >>> opt.plot(call=False, figsize=(10,10))
        """

        if call == True:
            tree = self.call(tree=True)
        else:
            tree = self.put(tree=True)

        plotter = BiTreePlotter(tree)

        return plotter.plot(dx, dy, size, digits, **kwargs)


class JRBinomialTreeOption(CRRBinomialTreeOption):
    """
    There exist many extensions of the CRR model. Jarrow and Rudd (1983), JR, adjusted the CRR
    model to account for the local drift term. They constructed a binomial model where the first two
    moments of the discrete and continuous time return processes match. As a consequence a probabil-
    ity measure equal to one half results. Therefore the CRR and JR models are sometimes atrributed
    as equal jumps binomial trees and equal probabilities binomial trees.

    Parameters
    ----------
    S : float
        Level or index price.
    K : float
        Strike price.
    t : float
        Time-to-maturity in fractional years. i.e. 1/12 for 1 month, 1/252 for 1 business day, 1.0 for 1 year.
    r : float
        Risk-free-rate in decimal format (i.e. 0.01 for 1%).
    b : float
        Annualized cost-of-carry rate, e.g. 0.1 means 10%
    sigma : float
        Annualized volatility of the underlying asset. Optional if calculating implied volatility.
        Required otherwise. By default None.
    n : int
        Number of time steps to use. By default 5.

    Returns
    -------
    JRBinomialTreeOption object.

    Example
    -------
    >>> import finoptions as fo
    >>> opt = fo.binomial_tree_options.JRBinomialTreeOption(S=50, K=50, t=5/12, r=0.1, b=0., sigma=0.4)
    >>> opt.call()
    >>> opt.put()
    >>> opt.call(type='american')
    >>> opt.greeks(call=True)

    References
    ----------
    [1] Haug E.G., The Complete Guide to Option Pricing Formulas
    """

    __name__ = "JRBinomialTreeOption"
    __title__ = "JR Binomial Tree Model"

    def _calc_price(self, z, n, type, tree=False):
        dt = self._t / n
        u = _np.exp((self._b - self._sigma ** 2 / 2) * dt + self._sigma * _np.sqrt(dt))
        d = _np.exp((self._b - self._sigma ** 2 / 2) * dt - self._sigma * _np.sqrt(dt))
        p = 1 / 2
        Df = _np.exp(-self._r * dt)

        OptionValue = z * (
            self._S * u ** _np.arange(0, n + 1) * d ** _np.arange(n, -1, -1) - self._K
        )
        OptionValue = (_np.abs(OptionValue) + OptionValue) / 2

        if type == "european":
            out = self._euro(OptionValue, n, Df, p, tree)
        elif type == "american":
            out = self._amer(OptionValue, n, Df, p, self._K, d, self._S, u, z, tree)

        if tree == False:
            return out[0]
        else:
            return out


class TIANBinomialTreeOption(CRRBinomialTreeOption):
    """
    Tian (1993) suggested to match discrete and continuous local moments up to third order.
    Leisen and Reimer (1996) proved that the order of convergence in pricing European options for all
    three methods is equal to one, and thus the three models are equivalent.

    Parameters
    ----------
    S : float
        Level or index price.
    K : float
        Strike price.
    t : float
        Time-to-maturity in fractional years. i.e. 1/12 for 1 month, 1/252 for 1 business day, 1.0 for 1 year.
    r : float
        Risk-free-rate in decimal format (i.e. 0.01 for 1%).
    b : float
        Annualized cost-of-carry rate, e.g. 0.1 means 10%
    sigma : float
        Annualized volatility of the underlying asset. Optional if calculating implied volatility.
        Required otherwise. By default None.
    type : str
        "european" to price European options, "american" to price American options. By default "european"
    n : int
        Number of time steps to use. By default 5.

    Returns
    -------
    JRBinomialTreeOption object.

    Example
    -------
    >>> import finoptions as fo
    >>> opt = fo.binomial_tree_options.TIANBinomialTreeOption(S=50, K=50, t=5/12, r=0.1, b=0., sigma=0.4)
    >>> opt.call()
    >>> opt.put()
    >>> opt.greeks(call=True)

    References
    ----------
    [1] Haug E.G., The Complete Guide to Option Pricing Formulas
    """

    def _calc_price(self, z, n, type, tree=False):
        dt = self._t / n
        M = _np.exp(self._b * dt)
        V = _np.exp(self._sigma ** 2 * dt)
        u = (M * V / 2) * (V + 1 + _np.sqrt(V * V + 2 * V - 3))
        d = (M * V / 2) * (V + 1 - _np.sqrt(V * V + 2 * V - 3))
        p = (M - d) / (u - d)
        Df = _np.exp(-self._r * dt)

        OptionValue = z * (
            self._S * u ** _np.arange(0, n + 1) * d ** _np.arange(n, -1, -1) - self._K
        )
        OptionValue = (_np.abs(OptionValue) + OptionValue) / 2

        if type == "european":
            out = self._euro(OptionValue, n, Df, p, tree)
        elif type == "american":
            out = self._amer(OptionValue, n, Df, p, self._K, d, self._S, u, z, tree)

        if tree == False:
            return out[0]
        else:
            return out


class TrinomialTreeOption(CRRBinomialTreeOption):
    """
    Calculates option prices from the Trinomial tree model. Trinomial trees in
    option pricing are similar to binomial trees. Trinomial trees can be used to
    price both European and American options on a single underlying asset.
    Because the asset price can move in three directions from a given node,
    compared with only two in a binomial tree, the number of time steps can
    be reduced to attain the same accuracy as in the binomial tree.

    Parameters
    ----------
    S : float
        Level or index price.
    K : float
        Strike price.
    t : float
        Time-to-maturity in fractional years. i.e. 1/12 for 1 month, 1/252 for 1 business day, 1.0 for 1 year.
    r : float
        Risk-free-rate in decimal format (i.e. 0.01 for 1%).
    b : float
        Annualized cost-of-carry rate, e.g. 0.1 means 10%
    sigma : float
        Annualized volatility of the underlying asset. Optional if calculating implied volatility.
        Required otherwise. By default None.
    type : str
        "european" to price European options, "american" to price American options. By default "european"
    n : int
        Number of time steps to use. By default 5.

    Returns
    -------
    JRBinomialTreeOption object.

    Example
    -------
    >>> import finoptions as fo
    >>> opt = fo.binomial_tree_options.TIANBinomialTreeOption(S=50, K=50, t=5/12, r=0.1, b=0., sigma=0.4)
    >>> opt.call()
    >>> opt.put()
    >>> opt.greeks(call=True)

        Reference:
          E.G Haug, The Complete Guide to Option Pricing Formulas
          Chapter 3.2
    """

    __name__ = "TrinomialTreeOption"
    __title__ = "Trinomial Tree Model"

    def _calc_price(self, z, n, type, tree):

        dt = self._t / n

        # Up-and-down jump sizes:
        u = _np.exp(self._sigma * _np.sqrt(2 * dt))
        d = _np.exp(-self._sigma * _np.sqrt(2 * dt))
        # Probabilities of going up and down:
        pu = (
            (_np.exp(self._b * dt / 2) - _np.exp(-self._sigma * _np.sqrt(dt / 2)))
            / (
                _np.exp(self._sigma * _np.sqrt(dt / 2))
                - _np.exp(-self._sigma * _np.sqrt(dt / 2))
            )
        ) ** 2
        pd = (
            (_np.exp(self._sigma * _np.sqrt(dt / 2)) - _np.exp(self._b * dt / 2))
            / (
                _np.exp(self._sigma * _np.sqrt(dt / 2))
                - _np.exp(-self._sigma * _np.sqrt(dt / 2))
            )
        ) ** 2

        # Probability of staying at the same asset price level:
        pm = 1 - pu - pd
        Df = _np.exp(-self._r * dt)

        # init tree as 1-D array
        OptionValue = _np.repeat(0.0, 2 * n + 1)
        for i in _np.arange(0, (2 * n + 1)):
            a = 0
            b = z * (
                self._S
                * u ** _np.maximum(i - n, 0)
                * d ** _np.maximum(n * 2 - n - i, 0)
                - self._K
            )
            OptionValue[i] = _np.maximum(a, b)

        list = []
        tr = _np.array(list)
        tr = OptionValue.copy()
        for j in _np.arange(0, n)[::-1]:
            for i in _np.arange(0, j * 2 + 1):
                # fmt: off
                if type == "european":
                    OptionValue[i] = self._euro(i, OptionValue, Df, pu, pd, pm)
                elif type == "american":
                    OptionValue[i] = self._amer(i, j, OptionValue, Df, pu, pd, pm, self._K, d, self._S, u, z)
                # fmt: on
                tr = _np.append(tr, OptionValue[i])

        if tree == True:
            tr = self._reshape(tr, n + 1)
            return tr
        else:
            return OptionValue[0]

    def _euro(self, i, OptionValue, Df, pu, pd, pm):
        # fmt: off
        out = (pu * OptionValue[i + 2] + pm * OptionValue[i + 1] + pd * OptionValue[i]) * Df
        # fmt: on
        return out

    def _amer(self, i, j, OptionValue, Df, pu, pd, pm, K, d, S, u, z):
        # fmt: off
        a = (z * (S * u ** _np.maximum(i - j, 0) * d ** _np.maximum(j - i, 0) - K))
        b = ( pu * OptionValue[i + 2] + pm * OptionValue[i + 1] + pd * OptionValue[i])* Df
        out = _np.maximum(a, b)
        # fmt: on

        return out

    @docstring_from(GreeksFDM.delta)
    def delta(self, call: bool = True):
        return self._greeks.delta(call=call)

    @docstring_from(GreeksFDM.theta)
    def theta(self, call: bool = True):
        return self._greeks.theta(call=call)

    @docstring_from(GreeksFDM.vega)
    def vega(self):
        return self._greeks.vega()

    @docstring_from(GreeksFDM.rho)
    def rho(self, call: bool = True):
        return self._greeks.rho(call=call)

    @docstring_from(GreeksFDM.lamb)
    def lamb(self, call: bool = True):
        return self._greeks.lamb(call=call)

    @docstring_from(GreeksFDM.gamma)
    def gamma(self):
        return self._greeks.gamma()

    @docstring_from(GreeksFDM.greeks)
    def greeks(self, call: bool = True):
        return self._greeks.greeks(call=call)

    def _reshape(self, tr, n):
        out = _np.zeros((2 * n - 1, n))
        for i in range(0, out.shape[1]):
            if i > 0:
                out[n - i - 1 : n + i, i] = tr[::-1][i ** 2 : i ** 2 + 2 * i + 1]
            else:
                out[n - 1][i] = tr[::-1][0]

        return out

    def plot(self, call=True, dx=-0.2, dy=0.2, size=12, digits=2, **kwargs):
        """
        Method to plot the trinomial tree values

        Parameters
        ----------
        call : bool
            Set to True to plot a call option. Otherwise set to False for a put option. By default True.
        dx : float
            x-offset for the node values in the same units at the call or put option value.
        dy : float
            y-offset for the node values in the same units at the call or put option value.
        size : float
            Font size for node and axis labels. By default 12. Title is 30% larger.
        digits : int
            Number of significant figures to show for option values. By default 2.
        **kwargs
            paramters to pass to matplotlib

        Returns
        -------
        matplotlib figure object

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.binomial_tree_options.CRRBinomialTreeOption(S=50, K=50, t=5/12, r=0.1, b=0.1, sigma=0.4, n=5, type='american')
        >>> opt.plot(call=False, figsize=(10,10))
        """

        if call == True:
            tree = self.call(tree=True)
        else:
            tree = self.put(tree=True)

        plotter = TriTreePlotter(tree)

        return plotter.plot(dx, dy, size, digits, **kwargs)



import numpy as _np
from ..base import GreeksFDM, Option as _Option
from ..utils import docstring_from


class BionomialSpreadAllTypes(_Option):
    """
    Rubinstein (1994) published a method to construct a 3-dimensional binomial 
    model that can be used to price most types of options that depend on two 
    assets - both American and European.

    Notes
    -----
    This model includes a cost of carry term b, the model can
    used to price European and American Options on:

    b = r       stocks
    b = r - q   stocks and stock indexes paying a continuous dividend yield q
    b = 0       futures
    b = r - rf  currency options with foreign interst rate rf

    Parameters
    ----------
    S1 : float
        Level or index price of asset 1.
    S2 : float
        Level or index price of asset 2.
    t : float
        Time-to-maturity in fractional years. i.e. 1/12 for 1 month, 1/252 for 1 business day, 1.0 for 1 year.
    r : float
        Risk-free-rate in decimal format (i.e. 0.01 for 1%).
    b1 : float
        Annualized cost-of-carry rate for asset 1, e.g. 0.1 means 10%
    b2 : float
        Annualized cost-of-carry rate for asset 2, e.g. 0.1 means 10%
    sigma1 : float
        Annualized volatility of the underlying asset 1. Optional if calculating implied volatility.
        Required otherwise. By default None.
    sigma2 : float
        Annualized volatility of the underlying asset 2. Optional if calculating implied volatility.
        Required otherwise. By default None.
    rho : float
        Correlation between asset 1 and asset 2.
    K : float
        Strike price. By default None.
    K2 : float
        Strike price. By default None.
    Q1 : float
        Weighting factor for asset 1 for use in payoff formula. By default 1.
    Q2 : float
        Weighting factor for asset 2 for use in payoff formula. By default 1.
    option : int
        Used to select payoff function for different spread option types.
        By default 1 for a spread option.

        option == 1:
            Spread Option
            C: max( 0, (Q1*S1 - Q2*S2) - K )
            P: max( 0, K - (Q1*S1 - Q2*S2) )
        option == 2:
            Options on the maximum
            C: max( 0, max(Q1*S1, Q2*S2) - K )
            P: max( 0, K - max(Q1*S1, Q2*S2) )
        option == 3:
            Options on the minimum
            C: max( 0, min(Q1*S1, Q2*S2) - K )
            P: max( 0, K - min(Q1*S1, Q2*S2) )
        option == 4:
            Dual-strike Option
            C: max( 0, (Q1*S1 - K), (Q2*S2 - K2) )
            P: max( 0, (K - Q1*S1), (K2 - Q2*S2) )
        option == 5:
            Reverse dual-strike option
            C: max( 0, (Q1*S1 - K), (K2 - Q2*S2) )
            P: max( 0, (K - Q1*S1), (Q2*S2 - K2) )
        option == 6:
            Portfolio options
            C: max( 0, (Q1*S1 + Q2*S2) - K )
            P: max( 0, K - (Q2*S2 + Q1*S1) )
        option == 7:
            Options to exchange one asset for another
            max( 0, Q2*S2 - Q1*S1 )
        option == 8:
            Relative performance options
            C: max( 0, (Q1*S1)/(Q2*S2) - K )
            P: max( 0, K - (Q1*S1)/(Q2*S2) )
        option == 9:
            Product options
            C: max( 0, (Q1*S1*Q2*S2) - K )
            P: max( 0, K - (Q1*S1*Q2*S2) )
    otype : str
        "european" to price European options, "american" to price American options. By default "european"
    n : int
        Number of time steps to use. By default 5.

    Returns
    -------
    BinomialSpreadOption object.

    Example
    -------
    >>> import finoptions as fo
    >>> opt = fo.spread_options.BinomialSpreadOption(S1=122, S2=120, K=3, r=0.1, b1=0, b2=0, n=100, 
            sigma1=0.2, sigma2=0.2, t=0.1, rho=-0.5, otype="european")
    >>> opt.call()
    >>> opt.put()
    >>> opt.greeks(call=True)

    References
    ----------
    [1] Haug E.G., The Complete Guide to Option Pricing Formulas
    """
    __name__ = "BionomialSpreadOption"
    __title__ = "Binomial Tree Spread Option Model"

    def __init__(
        self,
        S1: float,
        S2: float,
        t: float,
        r: float,
        b1: float,
        b2: float,
        sigma1: float,
        sigma2: float,
        rho: float,
        K: float = None,
        K2: float = None,
        Q1: float = 1,
        Q2: float = 1,
        option: int = 1,
        otype: str = "european",
        n: int = 5,
    ):
        self._S1 = S1
        self._S2 = S2
        self._Q1 = Q1
        self._Q2 = Q2
        self._K = K
        self._K2 = K2
        self._t = t
        self._r = r
        self._b1 = b1
        self._b2 = b2
        self._sigma1 = sigma1
        self._sigma2 = sigma2
        self._rho = rho
        self._option = option
        self._n = n
        self._otype = otype

        self._dt = self._t / self._n
        self._mu1 = self._b1 - 0.5 * self._sigma1 * self._sigma1
        self._mu2 = self._b2 - 0.5 * self._sigma2 * self._sigma2

        self._u = _np.exp(self._mu1 * self._dt + self._sigma1 * _np.sqrt(self._dt))
        self._d = _np.exp(self._mu1 * self._dt - self._sigma1 * _np.sqrt(self._dt))

        self._greeks = GreeksFDM(self)

    def _calc_price(self, z, n, otype, tree):
        # fmt: off
        
        OptionValue = _np.zeros((n+2,n+2))

        for j in range(0,n+1):
            Y1 = (2*j-n)*_np.sqrt(self._dt)
            node_value_S1 = self._S1*self._u**j*self._d**(n-j)

            for i in range(0,n+1):
                node_value_S2 = self._S2*_np.exp(self._mu2*n*self._dt) \
                    *_np.exp(self._sigma2*(self._rho*Y1 + _np.sqrt(1-self._rho**2)* (2*i-n)*_np.sqrt(self._dt)))
                OptionValue[j,i] = self._payoff(self._option, z, node_value_S1, node_value_S2, self._Q1, self._Q2, self._K, self._K2)

        for m in range(0,n)[::-1]:
            
            for j in range(0,m+1):
                Y1 = (2*j-m)*_np.sqrt(self._dt)
                node_value_S1 = self._S1*self._u**j*self._d**(m-j)
                for i in range(0,m+1):
                    y2 = self._rho * Y1 + _np.sqrt(1-self._rho**2) * (2*i-m) * _np.sqrt(self._dt)
                    node_value_S2 = self._S2 * _np.exp(self._mu2*m*self._dt)*_np.exp(self._sigma2*y2)
                    OptionValue[j,i] = 0.25*(OptionValue[j,i]
                        + OptionValue[j+1,i] + OptionValue[j,i+1]
                        + OptionValue[j+1,i+1]
                    )*_np.exp(-self._r*self._dt)

                    if otype == "american":
                        OptionValue[j,i] = _np.maximum(OptionValue[j,i], 
                                self._payoff(self._option, z, node_value_S1, node_value_S2, self._Q1, self._Q2, self._K, self._K2)
                        )


        if tree == True:
            out = OptionValue
        else:
            out = OptionValue[0,0]

        return out
    
    def _payoff(self, option, z, S1, S2, Q1, Q2, K, K2):
        
        if option == 1: # spread option
            return _np.maximum(0, z*((Q1*S1 - Q2*S2) - K))
        elif option == 2: # Max of two assets option
            return _np.maximum(0, z*(_np.maximum(Q1*S1, Q2*S2) - K))
        elif option == 3: # Min of two assets option
            return _np.maximum(0, z*(_np.minimum(Q1*S1, Q2*S2) - K))
        elif option == 4: # Dual strike option
            p1 = _np.maximum(0, z*(Q1*S1 - K))
            p2 = _np.maximum(0, z*(Q2*S2 - K2))
            return  _np.maximum(p1,p2) 
        elif option == 5: # Reverse-dual strike option
            p1 = _np.maximum(0, z*(Q1*S1 - K))
            p2 = _np.maximum(0, z*(K2 - Q2*S2))
            return  _np.maximum(p1,p2) 
        elif option == 6: # Portfolio option
            return _np.maximum(0, z*((Q1*S1 + Q2*S2) - K))
        elif option == 7: # Exchange option
            return _np.maximum(0, z*(Q2*S2 - Q1*S1))
        elif option == 8: # Outperformance option
            return _np.maximum(0, z*((Q1*S1 / Q2*S2) - K))
        elif option == 9: # Product option
            return _np.maximum(0, z*((Q1*S1 * Q2*S2) - K))
        
    def call(self, tree: bool = False):
        """
        Returns the calculated price of a call option according to the
        Rubinstein 3D Binomial Tree option price model.

        Parameters
        ----------
        tree : bool
            If True, returns the bionomial option as a tree-matrix to show the evolution of the option
            value.

        Returns
        -------
        float or tree-matrix

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.spread_options.BinomialSpreadOption(S1=122, S2=120, K=3, r=0.1, b1=0, b2=0, n=100, 
            sigma1=0.2, sigma2=0.2, t=0.1, rho=-0.5, otype="european")
        >>> opt.call()

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        z = 1
        return self._calc_price(z, self._n, self._otype, tree)

    def put(self, tree: bool = False):
        """
        Returns the calculated price of a put option according to the
        Rubinstein 3D Binomial Tree option price model.

        Parameters
        ----------
        tree : bool
            If True, returns the bionomial option as a tree-matrix to show the evolution of the option
            value.

        Returns
        -------
        float or tree-matrix

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.spread_options.BinomialSpreadOption(S1=122, S2=120, K=3, r=0.1, b1=0, b2=0, n=100, 
            sigma1=0.2, sigma2=0.2, t=0.1, rho=-0.5, otype="european")
        >>> opt.put()

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        z = -1
        return self._calc_price(z, self._n, self._otype, tree)

    @docstring_from(GreeksFDM.delta)
    def delta(self, call: bool = True):
        
        fd1 = self._greeks._make_partial_der("S1", call, self, n=1)
        fd2 = self._greeks._make_partial_der("S2", call, self, n=1)

        # multiple by 1 to return float vs array for single element arrays. Multi-element arrays returned as normal
        out = dict(
            S1 = fd1(self._S1) * 1,
            S2 = fd2(self._S2) * 1,
        )

        return out

    @docstring_from(GreeksFDM.theta)
    def theta(self, call: bool = True):
        return self._greeks.theta(call=call)

    @docstring_from(GreeksFDM.vega)
    def vega(self):
        # same for both call and put options
        fd1 = self._greeks._make_partial_der("sigma1", True, self, n=1)
        fd2 = self._greeks._make_partial_der("sigma2", True, self, n=1)

        out = dict(
            sigma1 = fd1(self._sigma1) * 1,
            sigma2 = fd2(self._sigma2) * 1
        )

        # multiple by 1 to return float vs array for single element arrays. Multi-element arrays returned as normal
        return out

    @docstring_from(GreeksFDM.rho)
    def rho(self, call: bool = True):
        return self._greeks.rho(call=call)

    @docstring_from(GreeksFDM.lamb)
    def lamb(self, call: bool = True):
        if call == True:
            price = self.call()
        else:
            price = self.put()

        out = dict(
            S1 = self.delta(call=call)['S1'] * self._S1 / price,
            S2 = self.delta(call=call)['S2'] * self._S2 / price
        )
        return out

    @docstring_from(GreeksFDM.gamma)
    def gamma(self):
        # same for both call and put options
        fd1 = self._greeks._make_partial_der("S1", True, self, n=2)
        fd2 = self._greeks._make_partial_der("S2", True, self, n=2)
        
        out = dict(
            S1 = fd1(self._S1) * 1,
            S2 = fd2(self._S2) * 1
        )
        # multiple by 1 to return float vs array for single element arrays. Multi-element arrays returned as normal
        return out

    @docstring_from(GreeksFDM.greeks)
    def greeks(self, call: bool = True):
        gk = {
            "delta": self.delta(call),
            "theta": self.theta(call),
            "vega": self.vega(),
            "rho": self.rho(call),
            "lambda": self.lamb(call),
            "gamma": self.gamma(),
        }

        return gk

    def get_params(self):
        return {
            "S1": self._S1,
            "S2": self._S2,
            "K": self._K,
            "K2": self._K2,
            "t": self._t,
            "r": self._r,
            "b1": self._b1,
            "b2": self._b2,
            "sigma1": self._sigma1,
            "sigma2": self._sigma2,
            "rho": self._rho,
            "Q1": self._Q1,
            "Q2": self._Q2,
            "option": self._option,
            "otype": self._otype,
            "n": self._n,
        }

    def summary(self, printer=True):
        """
        Print summary report of option

        Parameters
        ----------
        printer : bool
            True to print summary. False to return a string.
        """
        out = f"Title: {self.__title__} Valuation\n\nParameters:\n\n"

        params = self.get_params()

        for p in params:
            out += f"  {p} = {self._check_string(params[p])}\n"

        try:
            # if self._sigma or its variations are not None add call and put prices
            if isinstance(self.call(), _np.ndarray):
                c = self._check_string(self.call().round(2))
                p = self._check_string(self.put().round(2))
                price = f"\nOption Price:\n\n  call-{self._otype}: {c}\n  put-{self._otype}: {p}"
            else:
                price = f"\nOption Price:\n\n  call-{self._otype}: {round(self.call(),6)}\n  put-{self._otype}: {round(self.put(),6)}"
            out += price
        except:
            pass

        if printer == True:
            print(out)
        else:
            return out


class BionomialMaxOption(BionomialSpreadAllTypes):
    """
    Rubinstein (1994) published a method to construct a 3-dimensional binomial 
    model that can be used to price most types of options that depend on two 
    assets - both American and European.

    This implementation is for am option on the maximum and is defined as
    the maximum of S1 or S2 vs the strike price K.

    C: max( 0, max(Q1*S1, Q2*S2) - K )
    P: max( 0, K - max(Q1*S1, Q2*S2) )

    Notes
    -----
    This model includes a cost of carry term b, the model can
    used to price European and American Options on:

    b = r       stocks
    b = r - q   stocks and stock indexes paying a continuous dividend yield q
    b = 0       futures
    b = r - rf  currency options with foreign interst rate rf

    Parameters
    ----------
    S1 : float
        Level or index price of asset 1.
    S2 : float
        Level or index price of asset 2.
    K : float
        Strike price. By default None.
    t : float
        Time-to-maturity in fractional years. i.e. 1/12 for 1 month, 1/252 for 1 business day, 1.0 for 1 year.
    r : float
        Risk-free-rate in decimal format (i.e. 0.01 for 1%).
    b1 : float
        Annualized cost-of-carry rate for asset 1, e.g. 0.1 means 10%
    b2 : float
        Annualized cost-of-carry rate for asset 2, e.g. 0.1 means 10%
    sigma1 : float
        Annualized volatility of the underlying asset 1. Optional if calculating implied volatility.
        Required otherwise. By default None.
    sigma2 : float
        Annualized volatility of the underlying asset 2. Optional if calculating implied volatility.
        Required otherwise. By default None.
    rho : float
        Correlation between asset 1 and asset 2.
    Q1 : float
        Weighting factor for asset 1 for use in payoff formula. By default 1.
    Q2 : float
        Weighting factor for asset 2 for use in payoff formula. By default 1.
    otype : str
        "european" to price European options, "american" to price American options. By default "european"
    n : int
        Number of time steps to use. By default 5.

    Returns
    -------
    BinomialSpreadOption object.

    Example
    -------
    >>> import finoptions as fo
    >>> opt = fo.spread_options.BinomialSpreadOption(S1=122, S2=120, K=3, r=0.1, b1=0, b2=0, n=100, 
            sigma1=0.2, sigma2=0.2, t=0.1, rho=-0.5, otype="european")
    >>> opt.call()
    >>> opt.put()
    >>> opt.greeks(call=True)
    """
    __name__ = "BionomialMaxOption"
    __title__ = "Binomial Tree Maximum Spread Option Model"

    def __init__(
        self,
        S1: float,
        S2: float,
        K: float,
        t: float,
        r: float,
        b1: float,
        b2: float,
        sigma1: float,
        sigma2: float,
        rho: float,
        Q1: float = 1,
        Q2: float = 1,
        otype: str = "european",
        n: int = 5,
    ):
        self._S1 = S1
        self._S2 = S2
        self._Q1 = Q1
        self._Q2 = Q2
        self._K = K
        self._K2 = None
        self._t = t
        self._r = r
        self._b1 = b1
        self._b2 = b2
        self._sigma1 = sigma1
        self._sigma2 = sigma2
        self._rho = rho
        self._option = 2
        self._n = n
        self._otype = otype

        self._dt = self._t / self._n
        self._mu1 = self._b1 - 0.5 * self._sigma1 * self._sigma1
        self._mu2 = self._b2 - 0.5 * self._sigma2 * self._sigma2

        self._u = _np.exp(self._mu1 * self._dt + self._sigma1 * _np.sqrt(self._dt))
        self._d = _np.exp(self._mu1 * self._dt - self._sigma1 * _np.sqrt(self._dt))

        self._greeks = GreeksFDM(self)

    def get_params(self):
        # need to override so that K2 and option isn't returned
        return {
            "S1": self._S1,
            "S2": self._S2,
            "K": self._K,
            "t": self._t,
            "r": self._r,
            "b1": self._b1,
            "b2": self._b2,
            "sigma1": self._sigma1,
            "sigma2": self._sigma2,
            "rho": self._rho,
            "Q1": self._Q1,
            "Q2": self._Q2,
            "otype": self._otype,
            "n": self._n,
        }


class BionomialMinOption(BionomialSpreadAllTypes):
    """
    Rubinstein (1994) published a method to construct a 3-dimensional binomial 
    model that can be used to price most types of options that depend on two 
    assets - both American and European.

    This implementation is for am option on the minimum and is defined as
    the maximum of S1 or S2 vs the strike price K.

    C: max( 0, min(Q1*S1, Q2*S2) - K )
    P: max( 0, K - min(Q1*S1, Q2*S2) )

    Notes
    -----
    This model includes a cost of carry term b, the model can
    used to price European and American Options on:

    b = r       stocks
    b = r - q   stocks and stock indexes paying a continuous dividend yield q
    b = 0       futures
    b = r - rf  currency options with foreign interst rate rf

    Parameters
    ----------
    S1 : float
        Level or index price of asset 1.
    S2 : float
        Level or index price of asset 2.
    K : float
        Strike price. By default None.
    t : float
        Time-to-maturity in fractional years. i.e. 1/12 for 1 month, 1/252 for 1 business day, 1.0 for 1 year.
    r : float
        Risk-free-rate in decimal format (i.e. 0.01 for 1%).
    b1 : float
        Annualized cost-of-carry rate for asset 1, e.g. 0.1 means 10%
    b2 : float
        Annualized cost-of-carry rate for asset 2, e.g. 0.1 means 10%
    sigma1 : float
        Annualized volatility of the underlying asset 1. Optional if calculating implied volatility.
        Required otherwise. By default None.
    sigma2 : float
        Annualized volatility of the underlying asset 2. Optional if calculating implied volatility.
        Required otherwise. By default None.
    rho : float
        Correlation between asset 1 and asset 2.
    Q1 : float
        Weighting factor for asset 1 for use in payoff formula. By default 1.
    Q2 : float
        Weighting factor for asset 2 for use in payoff formula. By default 1.
    otype : str
        "european" to price European options, "american" to price American options. By default "european"
    n : int
        Number of time steps to use. By default 5.

    Returns
    -------
    BinomialSpreadOption object.

    Example
    -------
    >>> import finoptions as fo
    >>> opt = fo.spread_options.BinomialSpreadOption(S1=122, S2=120, K=3, r=0.1, b1=0, b2=0, n=100, 
            sigma1=0.2, sigma2=0.2, t=0.1, rho=-0.5, otype="european")
    >>> opt.call()
    >>> opt.put()
    >>> opt.greeks(call=True)
    """
    __name__ = "BionomialMinOption"
    __title__ = "Binomial Tree Minimum Spread Option Model"

    def __init__(
        self,
        S1: float,
        S2: float,
        K: float,
        t: float,
        r: float,
        b1: float,
        b2: float,
        sigma1: float,
        sigma2: float,
        rho: float,
        Q1: float = 1,
        Q2: float = 1,
        otype: str = "european",
        n: int = 5,
    ):
        self._S1 = S1
        self._S2 = S2
        self._Q1 = Q1
        self._Q2 = Q2
        self._K = K
        self._K2 = None
        self._t = t
        self._r = r
        self._b1 = b1
        self._b2 = b2
        self._sigma1 = sigma1
        self._sigma2 = sigma2
        self._rho = rho
        self._option = 3
        self._n = n
        self._otype = otype

        self._dt = self._t / self._n
        self._mu1 = self._b1 - 0.5 * self._sigma1 * self._sigma1
        self._mu2 = self._b2 - 0.5 * self._sigma2 * self._sigma2

        self._u = _np.exp(self._mu1 * self._dt + self._sigma1 * _np.sqrt(self._dt))
        self._d = _np.exp(self._mu1 * self._dt - self._sigma1 * _np.sqrt(self._dt))

        self._greeks = GreeksFDM(self)
        
    def get_params(self):
        # need to override so that K2 and option isn't returned
        return {
            "S1": self._S1,
            "S2": self._S2,
            "K": self._K,
            "t": self._t,
            "r": self._r,
            "b1": self._b1,
            "b2": self._b2,
            "sigma1": self._sigma1,
            "sigma2": self._sigma2,
            "rho": self._rho,
            "Q1": self._Q1,
            "Q2": self._Q2,
            "otype": self._otype,
            "n": self._n,
        }


class BionomialSpreadOption(BionomialSpreadAllTypes):
    """
    Rubinstein (1994) published a method to construct a 3-dimensional binomial 
    model that can be used to price most types of options that depend on two 
    assets - both American and European.

    This implementation is for am option on the minimum and is defined as
    the maximum of S1 or S2 vs the strike price K.

    C: max( 0, (Q1*S1 - Q2*S2) - K )
    P: max( 0, K - (Q1*S1 - Q2*S2) )

    Notes
    -----
    This model includes a cost of carry term b, the model can
    used to price European and American Options on:

    b = r       stocks
    b = r - q   stocks and stock indexes paying a continuous dividend yield q
    b = 0       futures
    b = r - rf  currency options with foreign interst rate rf

    Parameters
    ----------
    S1 : float
        Level or index price of asset 1.
    S2 : float
        Level or index price of asset 2.
    K : float
        Strike price. By default None.
    t : float
        Time-to-maturity in fractional years. i.e. 1/12 for 1 month, 1/252 for 1 business day, 1.0 for 1 year.
    r : float
        Risk-free-rate in decimal format (i.e. 0.01 for 1%).
    b1 : float
        Annualized cost-of-carry rate for asset 1, e.g. 0.1 means 10%
    b2 : float
        Annualized cost-of-carry rate for asset 2, e.g. 0.1 means 10%
    sigma1 : float
        Annualized volatility of the underlying asset 1. Optional if calculating implied volatility.
        Required otherwise. By default None.
    sigma2 : float
        Annualized volatility of the underlying asset 2. Optional if calculating implied volatility.
        Required otherwise. By default None.
    rho : float
        Correlation between asset 1 and asset 2.
    Q1 : float
        Weighting factor for asset 1 for use in payoff formula. By default 1.
    Q2 : float
        Weighting factor for asset 2 for use in payoff formula. By default 1.
    otype : str
        "european" to price European options, "american" to price American options. By default "european"
    n : int
        Number of time steps to use. By default 5.

    Returns
    -------
    BinomialSpreadOption object.

    Example
    -------
    >>> import finoptions as fo
    >>> opt = fo.spread_options.BinomialSpreadOption(S1=122, S2=120, K=3, r=0.1, b1=0, b2=0, n=100, 
            sigma1=0.2, sigma2=0.2, t=0.1, rho=-0.5, otype="european")
    >>> opt.call()
    >>> opt.put()
    >>> opt.greeks(call=True)
    """
    __name__ = "BionomialMinOption"
    __title__ = "Binomial Tree Minimum Spread Option Model"

    def __init__(
        self,
        S1: float,
        S2: float,
        K: float,
        t: float,
        r: float,
        b1: float,
        b2: float,
        sigma1: float,
        sigma2: float,
        rho: float,
        Q1: float = 1,
        Q2: float = 1,
        otype: str = "european",
        n: int = 5,
    ):
        self._S1 = S1
        self._S2 = S2
        self._Q1 = Q1
        self._Q2 = Q2
        self._K = K
        self._K2 = None
        self._t = t
        self._r = r
        self._b1 = b1
        self._b2 = b2
        self._sigma1 = sigma1
        self._sigma2 = sigma2
        self._rho = rho
        self._option = 1
        self._n = n
        self._otype = otype

        self._dt = self._t / self._n
        self._mu1 = self._b1 - 0.5 * self._sigma1 * self._sigma1
        self._mu2 = self._b2 - 0.5 * self._sigma2 * self._sigma2

        self._u = _np.exp(self._mu1 * self._dt + self._sigma1 * _np.sqrt(self._dt))
        self._d = _np.exp(self._mu1 * self._dt - self._sigma1 * _np.sqrt(self._dt))

        self._greeks = GreeksFDM(self)

    def get_params(self):
        # need to override so that K2 and option isn't returned
        return {
            "S1": self._S1,
            "S2": self._S2,
            "K": self._K,
            "t": self._t,
            "r": self._r,
            "b1": self._b1,
            "b2": self._b2,
            "sigma1": self._sigma1,
            "sigma2": self._sigma2,
            "rho": self._rho,
            "Q1": self._Q1,
            "Q2": self._Q2,
            "otype": self._otype,
            "n": self._n,
        }


from abc import ABC, abstractmethod
import numpy as _np
from math import atan as _atan
import copy as _copy
import numdifftools as _nd
import sys as _sys
import warnings as _warnings
from scipy.optimize import root_scalar as _root_scalar
from scipy.optimize import root as _root
import scipy.optimize as _opt


class NormalDistFunctions:
    def _NDF(self, x):
        """
        Calculate the normal distribution function of x

        Parameters
        ----------

        x : float
            Value to calculate the normal distribution function for x.

        Returns
        -------
        float

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """

        return _np.exp(-x * x / 2) / _np.sqrt(8 * _atan(1))

    def _CND(self, x):
        """
        Calculate the cumulated normal distribution function of x

        Parameters
        ----------

        x : float
            Value to calculate the cumulated normal distribution function for x.

        Returns
        -------
        float

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """

        k = 1 / (1 + 0.2316419 * abs(x))
        a1 = 0.319381530
        a2 = -0.356563782
        a3 = 1.781477937
        a4 = -1.821255978
        a5 = 1.330274429
        result = (
            self._NDF(x)
            * (a1 * k + a2 * k ** 2 + a3 * k ** 3 + a4 * k ** 4 + a5 * k ** 5)
            - 0.5
        )
        result = 0.5 - result * _np.sign(x)

        return result

    def _CBND(self, x1, x2, rho):
        """
        Calculate the cumulative bivariate normal distribution function.

        Haug E.G., The Complete Guide to Option Pricing Formulas

        Compute:
        Take care for the limit rho = +/- 1
        """

        a = x1
        b = x2
        if abs(rho) == 1:
            rho = rho - (1e-12) * _np.sign(rho)
        # cat("\n a - b - rho :"); print(c(a,b,rho))
        X = [0.24840615, 0.39233107, 0.21141819, 0.03324666, 0.00082485334]
        y = [0.10024215, 0.48281397, 1.0609498, 1.7797294, 2.6697604]
        a1 = a / _np.sqrt(2 * (1 - rho ** 2))
        b1 = b / _np.sqrt(2 * (1 - rho ** 2))
        if (a <= 0) & (b <= 0) & (rho <= 0):
            Sum1 = 0
            for I in range(0, 5):
                for j in range(0, 5):
                    Sum1 = Sum1 + X[I] * X[j] * _np.exp(
                        a1 * (2 * y[I] - a1)
                        + b1 * (2 * y[j] - b1)
                        + 2 * rho * (y[I] - a1) * (y[j] - b1)
                    )
            result = _np.sqrt(1 - rho ** 2) / _np.pi * Sum1
            return result

        if (a <= 0) & (b >= 0) & (rho >= 0):
            result = self._CND(a) - self._CBND(a, -b, -rho)
            return result

        if (a >= 0) & (b <= 0) & (rho >= 0):
            result = self._CND(b) - self._CBND(-a, b, -rho)
            return result

        if (a >= 0) & (b >= 0) & (rho <= 0):
            result = self._CND(a) + self._CND(b) - 1 + self._CBND(-a, -b, rho)
            return result

        if (a * b * rho) >= 0:
            rho1 = (
                (rho * a - b)
                * _np.sign(a)
                / _np.sqrt(a ** 2 - 2 * rho * a * b + b ** 2)
            )
            rho2 = (
                (rho * b - a)
                * _np.sign(b)
                / _np.sqrt(a ** 2 - 2 * rho * a * b + b ** 2)
            )
            delta = (1 - _np.sign(a) * _np.sign(b)) / 4
            result = self._CBND(a, 0, rho1) + self._CBND(b, 0, rho2) - delta
            return result


class Display:
    """
    Class for printing summaries of Options classes

    Parameters
    ----------
    opt : Option class
        Option class with Option values to calculate greeks from
    call : bool
        True if to include call option price, False to exclude. By default True.
    put : bool
        True if to include put option price, False to exclude. By default True.
    """

    def __init__(self, opt, call=True, put=True):
        if isinstance(opt, Option):
            self._opt = opt
            self._call = call
            self._put = put
        else:
            raise ValueError("Parameter opt is not of the Option class")

    def _check_string(self, x):
        """
        helper function for summary method. Checks to see if the variable x is a numpy
        ndarray type and if it's length is greater than 6. If so, shortens the repsentation
        so that it fits.
        """
        if isinstance(x, _np.ndarray):
            if x.shape[0] > 6:
                return _np.array2string(x.round(2), threshold=6)

        return x

    def __str__(self):
        return self.summary(printer=False)

    def __repr__(self):
        out = f"{self.__name__}("
        params = self.get_params()

        for p in params:
            out = out + str(self._check_string(params[p])) + ", "

        # get rid of trailing comman and close pararenthesis
        out = out[:-2]
        out += ")"

        return out

    def summary(self, printer=True):
        """
        Print summary report of option

        Parameters
        ----------
        printer : bool
            True to print summary. False to return a string.
        """
        out = f"Title: {self._opt.__title__} Valuation\n\nParameters:\n\n"

        params = self._opt.get_params()

        for p in params:
            out += f"  {p} = {self._opt._check_string(params[p])}\n"

        try:
            # for printing arrays of calls and puts (converts to string first)
            if isinstance(self._opt.call(), _np.ndarray):
                price = f"\nOption Price:\n\n  "

                if self._call == True:
                    c = self._check_string(self._opt.call().round(2))
                    price += f"call: {c}\n  "
                if self._put == True:
                    p = self._check_string(self._opt.put().round(2))
                    price += f"put: {p}"
            else:
                price = f"\nOption Price:\n\n  "
                if self._call == True:
                    price += f"call: {round(self._opt.call(),6)}\n  "
                if self._put == True:
                    price += f"put: {round(self._opt.put(),6)}"
            out += price
        except:
            raise ValueError("call or put option methods failed")

        if printer == True:
            print(out)
        else:
            return out


class Option(ABC):
    """
    Base interface class for all options

    """

    @property
    @abstractmethod
    def __name__(self):
        # abstract attribute
        pass

    @property
    @abstractmethod
    def __title__(self):
        # abstract attribute
        pass

    # @abstractmethod
    # def __str__(self):
    #     pass

    # @abstractmethod
    # def __repr__(self):
    #     pass

    def __str__(self):
        return self.summary(printer=False)

    def __repr__(self):
        out = f"{self.__name__}("
        params = self.get_params()

        for p in params:
            out = out + str(self._check_string(params[p])) + ", "

        # get rid of trailing comman and close pararenthesis
        out = out[:-2]
        out += ")"

        return out

    def get_params(self):
        return {
            "S": self._S,
            "K": self._K,
            "t": self._t,
            "r": self._r,
            "b": self._b,
            "sigma": self._sigma,
        }

    def _check_sigma(self, func):
        if self._sigma is None:
            raise ValueError(f"sigma not defined. Required for {func}() method")

    def copy(self):
        return _copy.deepcopy(self)

    # @abstractmethod
    # def get_params(self):
    #     pass

    def set_param(self, x: str, value: float):
        """
        Method to change a parameter once the class hass been initiatized.

        Parameters
        ----------
        x : str
            Name of parameter to change
        value : float
            New value to give to parameter

        Returns
        -------
        None
        """

        tmp = self.get_params()
        tmp[x] = value

        self.__init__(**tmp)

    @abstractmethod
    def summary(self):
        pass

    @abstractmethod
    def delta(self):
        pass

    @abstractmethod
    def theta(self):
        pass

    @abstractmethod
    def vega(self):
        pass

    @abstractmethod
    def rho(self):
        pass

    @abstractmethod
    def lamb(self):
        pass

    @abstractmethod
    def gamma(self):
        pass

    @abstractmethod
    def volatility(self):
        pass


class ImpliedVolatility:
    """
    Class used to calculate implied volatility for Option classes

    Parameters
    ----------
    opt : Option class
        Option class with Option values to calculate greeks from

    """

    def __init__(self, opt):
        if isinstance(opt, Option):
            self._opt = opt
        else:
            raise ValueError("Parameter opt is not of the Option class")

    def _max_array(self, *args):
        """
        helper function to get largest ndarray. Assumes at
        least one array.
        """
        maxArray = _np.array([0])
        for a in args:
            if isinstance(a, _np.ndarray):
                if maxArray.size < a.size:
                    maxArray = a

        return maxArray

    def _func(sigma, obj, call, price):
        """
        helper function to be used for volatility root finding.
        """
        temp = obj.copy()
        temp.set_param("sigma", sigma)
        if call == True:
            return price - temp.call()
        else:
            return price - temp.put()

    def _check_array(self, *args):
        """
        helper function to return True if any args past are numpy ndarrays
        """
        for a in args:
            # allows size 1 arrays. somethings greeks return size 1 arrays
            if isinstance(a, _np.ndarray) and a.size > 1:
                return True

        return False

    def _volatility(
        self,
        price: float,
        call: bool = True,
        tol=_sys.float_info.epsilon,
        maxiter=10000,
        verbose=False,
        _func=_func,
    ):
        """
        Compute the implied volatility of the GBSOption.

        Parameters
        ----------
        price : float
            Current price of the option
        call : bool
            Returns implied volatility for call option if True, else returns implied volatility for put options. By default True.
        tol : float
            max tolerance to fit the price to. By default system tolerance.
        maxiter : int
            number of iterations to run to fit price.
        verbose : bool
            True to return full optimization details from root finder function. False to just return the implied volatility numbers.

        Returns
        -------
        float

        Example
        -------
        """
        if self._opt._sigma is not None:
            _warnings.warn("sigma is not None but calculating implied volatility.")

        # check to see if arrays were past vs scalars.
        if self._check_array(price, *self._opt.get_params().values()):
            # if arrays, use root function
            a = self._opt._max_array(price, *self._opt.get_params().values())
            if verbose == True:
                sol = _root(_func, args=(self._opt, call, price), x0=_np.ones_like(a))
            else:
                sol = _root(_func, args=(self._opt, call, price), x0=_np.ones_like(a)).x
        else:
            # if scalars use root_scalar function
            if verbose == True:
                sol = _root_scalar(
                    _func,
                    (self._opt, call, price),
                    bracket=[-10, 10],
                    xtol=tol,
                    maxiter=maxiter,
                )
                # sol = _opt.brentq(_func, 1, 10000, xtol=tol, maxiter=maxiter)
            else:
                sol = _root_scalar(
                    _func,
                    (self._opt, call, price),
                    bracket=[-10, 10],
                    xtol=tol,
                    maxiter=maxiter,
                ).root

        return sol


if __name__ == "__main__":

    class TestOpt(Option):
        __name__ = "test"
        __repr__ = "test"

    opt = TestOpt()



import numpy as _np

from finoptions.spread_options.bitree3d import BionomialSpreadAllTypes
from ..base import GreeksFDM, Option as _Option
from ..utils import docstring_from

class SpreadApproxOption(_Option):
    """
    Rubinstein (1994) published a method to construct a 3-dimensional binomial
    model that can be used to price most types of options that depend on two
    assets - both American and European.

    Notes
    -----
    This model includes a cost of carry term b, the model can
    used to price European and American Options on:

    b = r       stocks
    b = r - q   stocks and stock indexes paying a continuous dividend yield q
    b = 0       futures
    b = r - rf  currency options with foreign interst rate rf

    Parameters
    ----------
    S1 : float
        Level or index price of asset 1.
    S2 : float
        Level or index price of asset 2.
    t : float
        Time-to-maturity in fractional years. i.e. 1/12 for 1 month, 1/252 for 1 business day, 1.0 for 1 year.
    r : float
        Risk-free-rate in decimal format (i.e. 0.01 for 1%).
    b1 : float
        Annualized cost-of-carry rate for asset 1, e.g. 0.1 means 10%
    b2 : float
        Annualized cost-of-carry rate for asset 2, e.g. 0.1 means 10%
    sigma1 : float
        Annualized volatility of the underlying asset 1. Optional if calculating implied volatility.
        Required otherwise. By default None.
    sigma2 : float
        Annualized volatility of the underlying asset 2. Optional if calculating implied volatility.
        Required otherwise. By default None.
    rho : float
        Correlation between asset 1 and asset 2.
    K : float
        Strike price. By default None.
    K2 : float
        Strike price. By default None.
    Q1 : float
        Weighting factor for asset 1 for use in payoff formula. By default 1.
    Q2 : float
        Weighting factor for asset 2 for use in payoff formula. By default 1.
    otype : str
        "european" to price European options, "american" to price American options. By default "european"
    n : int
        Number of time steps to use. By default 5.

    Returns
    -------
    BinomialSpreadOption object.

    Example
    -------
    >>> import finoptions as fo
    >>> opt = fo.spread_options.SpreadApproxOption(S1=122, S2=120, K=3, r=0.1, b1=0, b2=0, sigma1=0.2, sigma2=0.25, rho=0.5)
    >>> opt.call()
    >>> opt.put()
    >>> opt.greeks(call=True)

    References
    ----------
    [1] Haug E.G., The Complete Guide to Option Pricing Formulas
    """

    __name__ = "SpreadApproxOption"
    __title__ = "Spread-Option Approximation Model"

    def __init__(
        self,
        S1: float,
        S2: float,
        K: float,
        t: float,
        r: float,
        b1: float,
        b2: float,
        sigma1: float,
        sigma2: float,
        rho: float,
        Q1: float = 1,
        Q2: float = 1,
    ):
        self._S1 = S1
        self._S2 = S2
        self._Q1 = Q1
        self._Q2 = Q2
        self._K = K
        self._t = t
        self._r = r
        self._b1 = b1
        self._b2 = b2
        self._sigma1 = sigma1
        self._sigma2 = sigma2
        self._rho = rho

        self._e1 = _np.exp((b1 - r) * t)
        self._e2 = _np.exp((b2 - r) * t)

        self._S = (Q1 * S1 * self._e1) / (Q2 * S2 * self._e2 + K * _np.exp(-r * t))

        F = (Q2 * S2 * self._e2) / (Q2 * S2 * self._e2 + K * _np.exp(-r * t))

        self._sigma = _np.sqrt(
            sigma1 ** 2 + (sigma2 * F) ** 2 - 2 * rho * sigma1 * sigma2 * F
        )

        self._d1 = (_np.log(self._S) + 0.5 * self._sigma ** 2 * t) / (
            self._sigma * _np.sqrt(t)
        )

        self._d2 = self._d1 - self._sigma * _np.sqrt(t)

        self._greeks = GreeksFDM(self)

    def call(self):
        """
        Returns the calculated price of a call option using the Spread
        Option Approximation model.

        Returns
        -------
        float

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.spread_options.SpreadApproxOption(S1=122, S2=120, K=3, r=0.1, b1=0, b2=0, sigma1=0.2, sigma2=0.25, rho=0.5)
        >>> opt.call()

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        # fmt: off
        result = (self._Q2*self._S2*self._e2 + self._K*_np.exp(-self._r*self._t)) \
                    * (self._S*self._CND(self._d1) - self._CND(self._d2))

        return result

    def put(self):
        """
        Returns the calculated price of a put option using the Spread
        Option Approximation model.

        Returns
        -------
        float

        Example
        -------
        >>> import finoptions as fo
        >>> opt = fo.spread_options.SpreadApproxOption(S1=122, S2=120, K=3, r=0.1, b1=0, b2=0, sigma1=0.2, sigma2=0.25, rho=0.5)
        >>> opt.put()

        References
        ----------
        [1] Haug E.G., The Complete Guide to Option Pricing Formulas
        """
        # fmt: off
        result = (self._Q2*self._S2*self._e2 + self._K*_np.exp(-self._r*self._t)) \
                    * (-self._S*self._CND(-self._d1) + self._CND(-self._d2))

        return result

    @docstring_from(GreeksFDM.delta)
    def delta(self, call: bool = True):
        
        fd1 = self._greeks._make_partial_der("S1", call, self, n=1)
        fd2 = self._greeks._make_partial_der("S2", call, self, n=1)

        # multiple by 1 to return float vs array for single element arrays. Multi-element arrays returned as normal
        out = dict(
            S1 = fd1(self._S1) * 1,
            S2 = fd2(self._S2) * 1,
        )

        return out

    @docstring_from(GreeksFDM.theta)
    def theta(self, call: bool = True):
        return self._greeks.theta(call=call)

    @docstring_from(GreeksFDM.vega)
    def vega(self):
        # same for both call and put options
        fd1 = self._greeks._make_partial_der("sigma1", True, self, n=1)
        fd2 = self._greeks._make_partial_der("sigma2", True, self, n=1)

        out = dict(
            sigma1 = fd1(self._sigma1) * 1,
            sigma2 = fd2(self._sigma2) * 1
        )

        # multiple by 1 to return float vs array for single element arrays. Multi-element arrays returned as normal
        return out

    @docstring_from(GreeksFDM.rho)
    def rho(self, call: bool = True):
        return self._greeks.rho(call=call)

    @docstring_from(GreeksFDM.lamb)
    def lamb(self, call: bool = True):
        if call == True:
            price = self.call()
        else:
            price = self.put()

        out = dict(
            S1 = self.delta(call=call)['S1'] * self._S1 / price,
            S2 = self.delta(call=call)['S2'] * self._S2 / price
        )
        return out

    @docstring_from(GreeksFDM.gamma)
    def gamma(self):
        # same for both call and put options
        fd1 = self._greeks._make_partial_der("S1", True, self, n=2)
        fd2 = self._greeks._make_partial_der("S2", True, self, n=2)
        
        out = dict(
            S1 = fd1(self._S1) * 1,
            S2 = fd2(self._S2) * 1
        )
        # multiple by 1 to return float vs array for single element arrays. Multi-element arrays returned as normal
        return out

    @docstring_from(GreeksFDM.greeks)
    def greeks(self, call: bool = True):
        gk = {
            "delta": self.delta(call),
            "theta": self.theta(call),
            "vega": self.vega(),
            "rho": self.rho(call),
            "lambda": self.lamb(call),
            "gamma": self.gamma(),
        }

        return gk

    def get_params(self):
        return {
            "S1": self._S1,
            "S2": self._S2,
            "K": self._K,
            "t": self._t,
            "r": self._r,
            "b1": self._b1,
            "b2": self._b2,
            "sigma1": self._sigma1,
            "sigma2": self._sigma2,
            "rho": self._rho,
            "Q1": self._Q1,
            "Q2": self._Q2,
        }

        


from abc import ABC, abstractmethod
import numpy as _np


class Path(ABC):
    def __init__(self, epsilon, sigma, dt, b):
        self.epsilon = epsilon
        self.sigma = sigma
        self.dt = dt
        self.b = b

    @abstractmethod
    def generate_path(self):
        pass


class WienerPath(Path):
    """
    Generate Wiener Paths for use with Monte Carlo Options

    Parameters
    ----------
    epsilon : numpy array
        epilson for paths as numpy array. Output from Innovation class sample_innovation function
    sigma : float
        annualized volatility sigma used to generate paths
    dt : float
        time step dt to generate paths
    b : float
        Annualized cost-of-carry rate, e.g. 0.1 means 10%
    """

    def generate_path(self, **kwargs):
        # fmt: off
        return (self.b - (self.sigma ** 2) / 2) * self.dt + self.sigma * _np.sqrt(self.dt)  * self.epsilon
        # fmt: on



from ..base import Option as _Option
from ..vanillaoptions import GreeksFDM as _GreeksFDM
import numpy as _np
from scipy.optimize import minimize as _minimize
from scipy.stats import norm as _norm
from dataclasses import dataclass
from scipy.integrate import quad as _quad
# from .hnGARCH import *

# def _HNGCharacteristics(lamb, omega, alpha, beta, gamma, S, K, t_in_days, r_daily, call=True):
#     """
#     Characteristics function for Heston Nandi Option
#     """

#     premium = HNGOption(TypeFlag, model, S, X, Time.inDays, r.daily)
#     delta = HNGGreeks("Delta", TypeFlag, model, S, X, Time.inDays, r.daily)
#     gamma = HNGGreeks("Gamma", TypeFlag, model, S, X, Time.inDays, r.daily)

#     # Return Value:
#     list(premium = premium, delta = delta, gamma = gamma)

_eps = 1.49e-8
low_lim = 0
high_lim = _np.inf

class HNGGreeks(_GreeksFDM):
    pass


class HestonNandiOption:  # _Option
    """
    Option class for the Heston-Nandi Garch Option Model.

    Parameters
    ----------
    S : float
        Level or index price.
    K : float
        Strike price.
    t : float
        Time-to-maturity in days. i.e. 1/252 for 1 business day, 5/252 for 1 week etc...
    r : float
        Daily rate of interest. i.e. 0.25/252 means about 0.001% per day.
    lamb : float

    omega : float
        The GARCH model parameter which specifies the constant coefficient of the variance
        equation.
    alpha : float
        The GARCH model parameter which specifies the autoregressive coefficient
    beta : float
        The GARCH model parameter which specifies the variance coefficient.
    gamma : float
        The GARCH model parameter which specifies the asymmetry coefficient.

    Notes
    -----


    Returns
    -------
    HestonNandiOption object.

    Example
    -------
    >>> import finoptions as fo
    >>> opt = fo.heston_nandi_options.HestonNandiOption(S=80, K=82, t=1/3, td=1/4, r=0.06, D=4, sigma=0.30)
    >>> opt.call()
    >>> opt.greeks(call=True)

    References
    ----------
    [1] Haug E.G., The Complete Guide to Option Pricing Formulas
    """

    __name__ = "HestonNandiOption"
    __title__ = "The Heston-Nandi Garch Option Model"

    def __init__(
        self,
        S: float,
        K: float,
        t: float,
        r: float,
        lamb: float = None,
        omega: float = None,
        alpha: float = None,
        beta: float = None,
        gamma: float = None,
        model=None,
    ):

        self._S = S
        self._K = K
        self._t = t
        self._r = r
        self._lamb = lamb
        self._omega = omega
        self._alpha = alpha
        self._beta = beta
        self._gamma = gamma

    def call(self):
        """
        # Integrate:
        call1 = integrate(.fstarHN, 0, Inf, const = 1, model = model,
            S = S, X = X, Time.inDays = Time.inDays, r.daily = r.daily)

        # Compute Call Price:
        call.price = S/2 + exp(-r.daily*Time.inDays) * call1$value -
            X * exp(-r.daily*Time.inDays) * ( 1/2 + call2$value )

        """
        # fmt: off
        call1 = _quad(
            _fHN, low_lim, high_lim, 
            args=(1, self._lamb, self._omega, self._alpha, self._beta, self._gamma, self._S, self._K, self._t, self._r),
            epsabs=_eps, epsrel=_eps,
            )
        call2 = _quad(
            _fHN, low_lim, high_lim, 
            args=(0, self._lamb, self._omega, self._alpha, self._beta, self._gamma, self._S, self._K, self._t, self._r),
            epsabs=_eps, epsrel=_eps,
            )
        # fmt: on

        # Compute Call Price:
        price = (
            self._S / 2
            + _np.exp(-self._r * self._t) * call1[0]
            - self._K * _np.exp(-self._r * self._t) * (1 / 2 + call2[0])
        )

        return price

    def put(self):
        return self.call() + self._K * _np.exp(-self._r * self._t) - self._S


    def delta(self, call=True):
        # Integrate:
        # fmt: off
        delta1 = _quad(
            _fdeltaHN, low_lim, high_lim, 
            args=(1, self._lamb, self._omega, self._alpha, self._beta, self._gamma, self._S, self._K, self._t, self._r),
            epsabs=_eps, epsrel=_eps,
            )
        
        delta2 = _quad(
            _fdeltaHN, low_lim, high_lim, 
            args=(0, self._lamb, self._omega, self._alpha, self._beta, self._gamma, self._S, self._K, self._t, self._r),
            epsabs=_eps, epsrel=_eps,
            )
       
        # Compute Call and Put Delta :
        greek = 1/2 + _np.exp(-self._r*self._t) * delta1[0] - self._K * _np.exp(-self._r*self._t) * delta2[0]
        if call==False:
            greek = greek - 1

        return greek

    def gamma(self):
        # Integrate:
        # fmt: off
        gamma1 = _quad(
            _fgammaHN, low_lim, high_lim, 
            args=(1, self._lamb, self._omega, self._alpha, self._beta, self._gamma, self._S, self._K, self._t, self._r),
            epsabs=_eps, epsrel=_eps,
            )
        
        gamma2 = _quad(
            _fgammaHN, low_lim, high_lim, 
            args=(0, self._lamb, self._omega, self._alpha, self._beta, self._gamma, self._S, self._K, self._t, self._r),
            epsabs=_eps, epsrel=_eps,
            )
       
        # Compute Call and Put Delta :
        greek = _np.exp(-self._r*self._t) * gamma1[0] - self._K * _np.exp(-self._r*self._t) * gamma2[0]

        return greek


def _fHN(phi, const, lamb, omega, alpha, beta, gamma, S, K, t, r, real=True):
    """
    real=True to return fstarHN, real=False to return fHN
    """

    # Internal Function:

    # Model Parameters:
    gamma = gamma + lamb + 1 / 2
    lamb = -1 / 2
    sigma2 = (omega + alpha) / (1 - beta - alpha * gamma ** 2)
    # Function to be integrated:
    cphi0 = phi * _np.array([1j])
    cphi = cphi0 + const
    a = cphi * r
    b = lamb * cphi + cphi * cphi / 2
    # fmt: off
    for i in range(1, t):
        a = a + cphi*r + b*omega - _np.log(1-2*alpha*b)/2
        b = cphi*(lamb+gamma) - gamma**2/2 + beta*b + 0.5*(cphi-gamma)**2/(1-2*alpha*b)

    if real == True:
        f = _np.real(_np.exp(-cphi0*_np.log(K)+cphi*_np.log(S)+a+b*sigma2 )/cphi0)/_np.pi
    else:
        f = _np.exp(-cphi0*_np.log(K)+cphi*_np.log(S)+a+b*sigma2 )/cphi0/_np.pi

    # Return Value:
    return f
    # fmt: on


def _fdeltaHN(phi, const, lamb, omega, alpha, beta, gamma, S, K, t, r):
    # Function to be integrated:
    cphi0 = phi * _np.array([1j])
    cphi = cphi0 + const
    fdelta = cphi * _fHN(phi, const, lamb, omega, alpha, beta, gamma, S, K, t, r, real=False) / S

    # Return Value:
    return _np.real(fdelta)


def _fgammaHN(phi, const, lamb, omega, alpha, beta, gamma, S, K, t, r):
    # Function to be integrated:
    cphi0 = phi * _np.array([1j])
    cphi = cphi0 + const
    fgamma = (
        cphi * (cphi - 1) * _fHN(phi, const, lamb, omega, alpha, beta, gamma, S, K, t, r, real=False) / S ** 2
    )

    # Return Value:
    return _np.real(fgamma)

@dataclass
class ParamFit:
    llhHNGarch: float
    z: _np.array
    h: _np.array

    def __sub__(self,b):
        return self.llhHNGarch - b.llhHNGarch


def _llhHNGarch(x0, trace, symmetric, rfr, x, ret_obj=False):

    h = x.copy()
    z = x.copy()

    lamb = x0[0]
    omega = x0[1]
    alpha = x0[2]
    beta = x0[3]
    gamma = x0[4]

    # Transform - to keep them between 0 and 1:
    omega = 1 / (1 + _np.exp(-omega))
    alpha = 1 / (1 + _np.exp(-alpha))
    beta = 1 / (1 + _np.exp(-beta))

    # Add gamma if selected:
    if ~symmetric:
        gam = gamma
    else:
        gam = 0

    # HN Garch Filter:
    h[0] = (omega + alpha) / (1 - alpha * gam * gam - beta)
    z[0] = (x[0] - rfr - lamb * h[0]) / _np.sqrt(h[0])

    # fmt: off
    for i in range(1,len(z)):
        h[i] = omega + alpha * ( z[i-1] - gam * _np.sqrt(h[i-1]) )**2 + beta * h[i-1]
        z[i] = ( x[i] - rfr - lamb*h[i] ) / _np.sqrt(h[i])
    # fmt: on

    # Calculate Log - Likelihood for Normal Distribution:
    llhHNGarch = -_np.sum(_np.log(_norm.pdf(z) / _np.sqrt(h)))
    if trace:
        print("Parameter Estimate\n")
        print(lamb, omega, alpha, beta, gam)

    if ret_obj:
        params = ParamFit(llhHNGarch, z, h)
    else:
        params = llhHNGarch

    # Return Value:
    return params


def hngarch_fit(x, lamb = -0.5, omega = None, alpha = None, beta = 0.1, gamma = 0, rf = 0, symmetric = True, trace = True, **kw_nlm):
    # A function implemented by Diethelm Wuertz

    # Description:
    #   Fits Heston-Nandi Garch(1,1) time series model

    # Parameters:
    if omega is None:
        omega = _np.var(x)
    if alpha is None:
        alpha = 0.1*omega

    rfr = rf
    gam = gamma

    # Continue:
    params = dict(lamb = lamb, omega = omega, alpha = alpha,
        beta = beta, gamma = gam, rf = rfr)

    # Transform Parameters and Calculate Start Parameters:
    par_omega = -_np.log((1-omega)/omega)  # for 2
    par_alpha = -_np.log((1-alpha)/alpha)  # for 3
    par_beta = -_np.log((1-beta)/beta)     # for 4
    par_start = [lamb, par_omega, par_alpha, par_beta]
    if ~symmetric:
        par_start.append(gam)

    # Initial Log Likelihood:
    opt = dict()
    opt['value'] = _llhHNGarch(x0 = par_start, trace = trace, symmetric = symmetric, rfr = rfr, x = x)
    opt['estimate'] = par_start
    
    if trace:
        print(lamb, omega, alpha, beta, gam)
        print(opt['value'])
    
    # Estimate Parameters:
    res = _minimize(_llhHNGarch, par_start, args=(trace, symmetric, rfr, x), method='L-BFGS-B', **kw_nlm)

    # Log-Likelihood:
    opt['minimum'] = -res.fun + len(x)*_np.sqrt(2*_np.pi)
    # opt['params'] = params
    # opt['symmetric'] = symmetric
    opt['estimate'] = res.x

    # LLH, h, and z for Final Estimates:
    final = _llhHNGarch(opt['estimate'], trace = False, symmetric=symmetric, rfr=rfr, x=x, ret_obj=True)
    opt['h'] = final.h
    opt['z'] = final.z

    # Backtransform Estimated parameters:
    lamb = opt['estimate'][0]
    omega = (1 / (1+_np.exp(-opt['estimate'][1])))
    opt['estimate'][1] = omega

    alpha = (1 / (1+_np.exp(-opt['estimate'][2])))
    opt['estimate'][2] = alpha

    beta = (1 / (1+_np.exp(-opt['estimate'][3])))
    opt['estimate'][3] = beta
    
    if symmetric: 
        opt['estimate'][4] = 0
    gam = opt['estimate'][4]

    # names(opt$estimate) = c("lambda", "omega", "alpha", "beta", "gamma")

    # Add to Output:
    opt['model'] = dict(lamb = lamb, omega = omega, alpha = alpha,
        beta = beta, gamma = gam, rf = rfr)
    # opt['x'] = x

    # Statistics - Printing:
    opt['persistence'] = beta + alpha*gam*gam
    opt['sigma2'] = ( omega + alpha ) / ( 1 - opt['persistence'] )

    # Print Estimated Parameters:
    if (trace):
        print(opt['estimate'])

    opt['title'] = "Heston-Nandi Garch Parameter Estimation"

    return opt


def hngarch_sim(lamb = -0.5, omega = None, alpha = None, beta = 0.1, gamma = 0, rf = 0, n = 500, n_start = 0, inno=None, inno_start=None):
    """
    Description:
    Simulates a HN-GARCH time series with user supplied innovations.

    Details:
      The function simulates a Heston Nandi Garch(1,1) process with
      structure parameters specified through the parameters 
      lambda, omega, alpha, beta, gamma, rf.
      The function returns the simulated time series points
      neglecting those from the first "start_innov" innovations.

    Parameters
    ----------
    lamb : float
    omega : float
    alpha : float
    beta : float
    gamma : float
    rf : float
    n : int
    n_start : int
    inno : 
    inno_start :

    Example:
      x = hngarch()
      plot(100*x, type="l", xlab="Day numbers",
        ylab="Daily Returns %", main="Heston Nandi GARCH")
      S0 = 1
      plot(S0*exp(cumsum(x)), type="l", xlab="Day Numbers",
        ylab="Daily Prices", main="Heston Nandi GARCH") }
    """
    # FUNCTION:

    # Innovations:
    if inno is None:
        inno = _np.random.normal(0, 1, n)
    if (inno_start is None) & (n_start > 0):
        inno_start = _np.random.normal(0, 1, n_start)

    if inno_start is not None:
        x = _np.concatenate((inno_start, inno))
    else:
        x = inno.copy()
    h = x.copy()
    Z = x.copy()
    
    nt = n_start + n

    # Recursion:
    h[0] = ( omega + alpha )/( 1 - alpha*gamma*gamma - beta )
    x[0] = rf + lamb*h[0] + _np.sqrt(h[0]) * Z[0]
    for i in range(1,nt):
        h[i] = omega + alpha*(Z[i-1] - gamma*_np.sqrt(h[i-1]))**2 + beta*h[i-1]
        x[i] = rf + lamb*h[i] + _np.sqrt(h[i]) * Z[i]

    # Series:
    x = x[(n_start):]

    # Return Value:
    return x


from abc import ABC, abstractmethod
import numpy as _np


class Payoff(ABC):
    def __init__(
        self, path, S: float, K: float, t: float, r: float, b: float, sigma: float
    ):
        self.path = path
        self.S = S
        self.K = K
        self.t = t
        self.r = r
        self.b = b
        self.sigma = sigma

    @abstractmethod
    def call(self):
        pass

    @abstractmethod
    def put(self):
        pass


class PlainVanillaPayoff(Payoff):
    def call(self):
        St = self.S * _np.exp(_np.sum(self.path.generate_path(), axis=1))
        return _np.exp(-self.r * self.t) * _np.maximum(St - self.K, 0)

    def put(self):
        St = self.S * _np.exp(_np.sum(self.path.generate_path(), axis=1))
        return _np.exp(-self.r * self.t) * _np.maximum(self.K - St, 0)


class ArithmeticAsianPayoff(Payoff):
    def call(self):
        Sm = self.S * _np.exp(_np.cumsum(self.path.generate_path(), axis=1))
        Sm = _np.mean(Sm, axis=1)
        return _np.exp(-self.r * self.t) * _np.maximum(Sm - self.K, 0)

    def put(self):
        Sm = self.S * _np.exp(_np.cumsum(self.path.generate_path(), axis=1))
        Sm = _np.mean(Sm, axis=1)
        return _np.exp(-self.r * self.t) * _np.maximum(self.K - Sm, 0)



from ..base import Option as _Option
from ..vanillaoptions import GreeksFDM as _GreeksFDM
import numpy as _np
from scipy.integrate import quad as _quad
from scipy.stats import norm as _norm

from dataclasses import dataclass


@dataclass
class ParamFit:
    llhHNGarch: float
    Z: _np.array
    h: _np.array


def _llhHNGarch(x, lamb, omega, alpha, beta, gamma, trace, symmetric, rfr):

    h = x
    Z = x

    # Transform - to keep them between 0 and 1:
    omega = 1 / (1 + _np.exp(-omega))
    alpha = 1 / (1 + _np.exp(-alpha))
    beta = 1 / (1 + _np.exp(-beta))

    # Add gamma if selected:
    if ~symmetric:
        gam = gamma
    else:
        gam = 0

    # HN Garch Filter:
    h[1] = (omega + alpha) / (1 - alpha * gam * gam - beta)
    Z[1] = (x[1] - rfr - lamb * h[1]) / _np.sqrt(h[1])
    # fmt: off
    for i in range(1,len(Z)):
        h[i] = omega + alpha * ( Z[i-1] - gam * _np.sqrt(h[i-1]) )**2 + beta * h[i-1]
        Z[i] = ( x[i] - rfr - lamb*h[i] ) / _np.sqrt(h[i])
    # fmt: on

    # Calculate Log - Likelihood for Normal Distribution:
    llhHNGarch = -_np.sum(_np.log(_norm.pdf(Z) / _np.sqrt(h)))
    if trace:
        print("Parameter Estimate\n")
        print(lamb, omega, alpha, beta, gam)

    params = ParamFit()
    params.llhHNGarch = llhHNGarch
    params.Z = Z
    params.h = h

    # Return Value:
    return params


class HNGarch:
    def __init__(self):
        pass
